# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pytest

from lightworks import (
    Analyzer,
    Parameter,
    PhotonicCircuit,
    PhotonNumberError,
    PostSelection,
    State,
    Unitary,
    convert,
    random_unitary,
)
from lightworks.emulator import Backend

BACKEND = Backend("permanent")


class TestAnalyzer:
    """
    Unit tests to check results produced by Analyzer object in the emulator.
    """

    def setup_method(self) -> None:
        """Create a non-lossy and a lossy circuit for use."""
        self.circuit = PhotonicCircuit(4)
        self.lossy_circuit = PhotonicCircuit(4)
        for i, m in enumerate([0, 2, 1, 2, 0, 1]):
            self.circuit.bs(m)
            self.circuit.ps(m, phi=i)
            self.circuit.bs(m)
            self.circuit.ps(m + 1, phi=3 * i)
            # lossy circuit
            self.lossy_circuit.bs(
                m, loss=convert.db_loss_to_decimal(1 + 0.2 * i)
            )
            self.lossy_circuit.ps(m, phi=i)
            self.lossy_circuit.bs(
                m, loss=convert.db_loss_to_decimal(0.6 + 0.1 * i)
            )
            self.lossy_circuit.ps(m + 1, phi=3 * i)

    def test_hom(self):
        """Checks basic hom and confirms probability of |2,0> is 0.5."""
        circuit = PhotonicCircuit(2)
        circuit.bs(0)
        analyzer = Analyzer(circuit, State([1, 1]))
        results = BACKEND.run(analyzer)[State([1, 1])]
        p = results[State([2, 0])]
        assert pytest.approx(p) == 0.5

    def test_known_result(self):
        """
        Builds a circuit which produces a known result and checks this is found
        at the output.
        """
        # Build circuit
        circuit = PhotonicCircuit(4)
        circuit.bs(1, reflectivity=0.6)
        circuit.mode_swaps({0: 1, 1: 0, 2: 3, 3: 2})
        circuit.bs(0, 3, reflectivity=0.3)
        circuit.bs(0)
        # And check output counts
        analyzer = Analyzer(circuit, State([1, 0, 0, 1]))
        results = BACKEND.run(analyzer)[State([1, 0, 0, 1])]
        assert pytest.approx(abs(results[State([0, 1, 1, 0])])) == 0.5

    def test_analyzer_basic(self):
        """Check analyzer result with basic circuit."""
        analyzer = Analyzer(self.circuit, State([1, 0, 1, 0]))
        results = BACKEND.run(analyzer)[State([1, 0, 1, 0])]
        p = results[State([0, 1, 0, 1])]
        assert pytest.approx(p, 1e-8) == 0.6331805740170607

    def test_analyzer_basic_2photons_in_mode(self):
        """Check analyzer result with basic circuit."""
        analyzer = Analyzer(self.circuit, State([2, 0, 0, 0]))
        results = BACKEND.run(analyzer)[State([2, 0, 0, 0])]
        p = results[State([0, 1, 0, 1])]
        assert pytest.approx(p, 1e-8) == 0.0022854516590

    def test_analyzer_complex(self):
        """Check analyzer result when using post-selection and heralding."""
        # Add heralding mode
        self.circuit.herald(3, 0)
        analyzer = Analyzer(self.circuit, State([1, 0, 1]))
        # Just heralding
        results = BACKEND.run(analyzer)
        p = results[State([1, 0, 1]), State([0, 1, 1])]
        assert pytest.approx(p, 1e-8) == 0.091713377373246
        # Heralding + post-selection
        analyzer.post_selection = lambda s: s[0] == 1
        results = BACKEND.run(analyzer)
        p = results[State([1, 0, 1]), State([1, 1, 0])]
        assert pytest.approx(p, 1e-8) == 0.002934140618653
        # Check performance metric
        assert pytest.approx(results.performance, 1e-8) == 0.03181835438235

    def test_analyzer_complex_lossy(self):
        """
        Check analyzer result when using post-selection and heralding with a
        lossy circuit.
        """
        # Add heralding mode
        self.lossy_circuit.herald(3, 0)
        analyzer = Analyzer(self.lossy_circuit, State([1, 0, 1]))
        # Just heralding
        results = BACKEND.run(analyzer)
        p = results[State([1, 0, 1]), State([0, 1, 0])]
        assert pytest.approx(p, 1e-8) == 0.062204471804458
        # Heralding + post-selection
        analyzer.post_selection = lambda s: s[0] == 0
        results = BACKEND.run(analyzer)
        p = results[State([1, 0, 1]), State([0, 0, 1])]
        assert pytest.approx(p, 1e-8) == 0.0202286624257920
        p = results[State([1, 0, 1]), State([0, 0, 0])]
        assert pytest.approx(p, 1e-8) == 0.6051457174354371
        # Check performance metric
        assert pytest.approx(results.performance, 1e-8) == 0.6893563871958014

    def test_analyzer_complex_lossy_added_circuit(self):
        """
        Check analyzer result when using post-selection and heralding with a
        lossy circuit which has been added to another circuit.
        """
        # Add heralding mode
        self.lossy_circuit.herald(3, 0)
        new_circ = PhotonicCircuit(
            self.lossy_circuit.n_modes - len(self.lossy_circuit.heralds.input)
        )
        new_circ.add(self.lossy_circuit)
        analyzer = Analyzer(new_circ, State([1, 0, 1]))
        # Just heralding
        results = BACKEND.run(analyzer)
        p = results[State([1, 0, 1]), State([0, 1, 0])]
        assert pytest.approx(p, 1e-8) == 0.062204471804458
        # Heralding + post-selection
        analyzer.post_selection = lambda s: s[0] == 0
        results = BACKEND.run(analyzer)
        p = results[State([1, 0, 1]), State([0, 0, 1])]
        assert pytest.approx(p, 1e-8) == 0.0202286624257920
        p = results[State([1, 0, 1]), State([0, 0, 0])]
        assert pytest.approx(p, 1e-8) == 0.6051457174354371
        # Check performance metric
        assert pytest.approx(results.performance, 1e-8) == 0.6893563871958014

    def test_analyzer_error_rate(self):
        """Check the calculated error rate is correct for a given situation."""
        expectations = {
            State([1, 0, 1, 0]): State([0, 1, 0, 1]),
            State([0, 1, 0, 1]): State([1, 0, 1, 0]),
        }
        analyzer = Analyzer(
            self.circuit,
            [State([1, 0, 1, 0]), State([0, 1, 0, 1])],
            expected=expectations,
        )
        results = BACKEND.run(analyzer)
        assert pytest.approx(results.error_rate, 1e-8) == 0.46523865112110574

    def test_analyzer_circuit_update(self):
        """Check analyzer result before and after a circuit is modified."""
        circuit = Unitary(random_unitary(4))
        # Create analyzer and get results
        analyzer = Analyzer(circuit, State([1, 0, 1, 0]))
        analyzer.post_selection = lambda s: s[0] == 1
        results = BACKEND.run(analyzer)
        p = results[State([1, 0, 1, 0]), State([1, 1, 0, 0])]
        # Update circuit and get results
        circuit.bs(0)
        results = BACKEND.run(analyzer)
        p2 = results[State([1, 0, 1, 0]), State([1, 1, 0, 0])]
        assert p != p2

    def test_analyzer_circuit_parameter_update(self):
        """
        Check analyzer result before and after a circuit parameters is
        modified.
        """
        param = Parameter(0.3)
        circuit = PhotonicCircuit(4)
        circuit.bs(0, reflectivity=param)
        circuit.bs(2, reflectivity=param)
        circuit.bs(1, reflectivity=param)
        # Create analyzer and get results
        analyzer = Analyzer(circuit, State([1, 0, 1, 0]))
        post_select = PostSelection()
        post_select.add(0, 1)
        analyzer.post_selection = post_select
        results = BACKEND.run(analyzer)
        p = results[State([1, 0, 1, 0]), State([1, 1, 0, 0])]
        # Update parameter and get results
        param.set(0.65)
        results = BACKEND.run(analyzer)
        p2 = results[State([1, 0, 1, 0]), State([1, 1, 0, 0])]
        assert p != p2

    def test_circuit_assignment(self):
        """
        Checks that an incorrect value cannot be assigned to the circuit
        attribute.
        """
        circuit = Unitary(random_unitary(4))
        analyzer = Analyzer(circuit, State([1, 0, 1, 0]))
        with pytest.raises(TypeError):
            analyzer.circuit = random_unitary(5)

    def test_ns_gate(self):
        """
        Checks case of NS gate which was previously found to cause a bug.
        """
        u_ns = np.array(
            [
                [1 - 2**0.5, 2**-0.25, (3 / (2**0.5) - 2) ** 0.5],
                [2**-0.25, 0.5, 0.5 - 2**-0.5],
                [(3 / (2**0.5) - 2) ** 0.5, 0.5 - 2**-0.5, 2**0.5 - 0.5],
            ]
        )
        ns_gate = Unitary(u_ns)
        ns_gate.herald(1, 1)
        ns_gate.herald(2, 0)
        in_state = State([1])
        # Run simulation
        analyzer = Analyzer(ns_gate, in_state)
        results = BACKEND.run(analyzer)
        # Validate success probability is 0.25
        assert results[in_state, in_state] == pytest.approx(0.25)

    @pytest.mark.parametrize("n_output", [0, 1, 2])
    def test_herald_equivalance(self, n_output):
        """
        Checks that heralded and non-heralded circuits produce the same results.
        """
        # Create heralded and versions of same circuit
        unitary = random_unitary(6)
        circuit = Unitary(unitary)
        circuit_herald = Unitary(unitary)
        circuit_herald.herald((2, 2), 0)
        circuit_herald.herald((1, 3), (1, n_output))
        # Simulate both with equivalent inputs
        sim = Analyzer(circuit, State([0, 1, 0, 1, 1, 0]))
        results = BACKEND.run(sim)
        sim_h = Analyzer(circuit_herald, State([0, 1, 1, 0]))
        results_h = BACKEND.run(sim_h)
        # Then check equivalence of results for all outputs
        for output in results_h.outputs:
            full_state = output[0:2] + State([0, n_output]) + output[2:]
            assert (
                pytest.approx(results_h[State([0, 1, 1, 0]), output])
                == results[State([0, 1, 0, 1, 1, 0]), full_state]
            )

    @pytest.mark.parametrize("n_output", [0, 1, 2])
    def test_herald_equivalance_lossy(self, n_output):
        """
        Checks that heralded and non-heralded circuits produce the same results
        when a lossy circuit is used.
        """
        # Create heralded and versions of same circuit
        unitary = random_unitary(6)
        circuit = Unitary(unitary)
        for i in range(6):
            circuit.loss(i, (i + 1) / 10)
        circuit_herald = Unitary(unitary)
        for i in range(6):
            circuit_herald.loss(i, (i + 1) / 10)
        circuit_herald.herald((2, 2), 0)
        circuit_herald.herald((1, 3), (1, n_output))
        # Simulate both with equivalent inputs
        sim = Analyzer(circuit, State([0, 1, 0, 1, 1, 0]))
        results = BACKEND.run(sim)
        sim_h = Analyzer(circuit_herald, State([0, 1, 1, 0]))
        results_h = BACKEND.run(sim_h)
        # Then check equivalence of results for all outputs
        for output in results_h.outputs:
            full_state = output[0:2] + State([0, n_output]) + output[2:]
            assert (
                pytest.approx(results_h[State([0, 1, 1, 0]), output])
                == results[State([0, 1, 0, 1, 1, 0]), full_state]
            )

    @pytest.mark.parametrize("n_output", [0, 1, 2])
    def test_herald_equivalance_grouped(self, n_output):
        """
        Checks that heralded and non-heralded circuits produce the same results
        when a grouped circuit is used.
        """
        # Create heralded and versions of same circuit
        unitary = random_unitary(6)
        circuit = Unitary(unitary)
        for i in range(6):
            circuit.loss(i, (i + 1) / 10)
        sub_circuit = Unitary(unitary)
        for i in range(6):
            sub_circuit.loss(i, (i + 1) / 10)
        sub_circuit.herald((2, 2), 0)
        sub_circuit.herald((1, 3), (1, n_output))
        circuit_herald = PhotonicCircuit(4)
        circuit_herald.add(sub_circuit)
        # Simulate both with equivalent inputs
        sim = Analyzer(circuit, State([0, 1, 0, 1, 1, 0]))
        results = BACKEND.run(sim)
        sim_h = Analyzer(circuit_herald, State([0, 1, 1, 0]))
        results_h = BACKEND.run(sim_h)
        # Then check equivalence of results for all outputs
        for output in results_h.outputs:
            full_state = output[0:2] + State([0, n_output]) + output[2:]
            assert (
                pytest.approx(results_h[State([0, 1, 1, 0]), output])
                == results[State([0, 1, 0, 1, 1, 0]), full_state]
            )

    def test_output_heralds_too_large(self):
        """
        Confirms a PhotonNumberError is raised when the number of output heralds
        is larger than the number of photons input into the system.
        """
        circuit = Unitary(random_unitary(4))
        circuit.herald(3, (0, 2))
        sim = Analyzer(circuit, State([1, 0, 0]))
        with pytest.raises(PhotonNumberError):
            sim._generate_task()



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A custom state datatype, which is created for storing annotated state details.
It is not intended that this class will be ordinarily accessible to users.
"""

from collections.abc import Iterator
from typing import Any, Union, overload

from lightworks.emulator.utils.exceptions import AnnotatedStateError
from lightworks.emulator.utils.state import annotated_state_to_string


class AnnotatedState:
    """
    Acts as a custom data state which enables fock states to be defined, with
    the main difference to the core State class being that here they are
    defined with a label to indicate photon indistinguishability.

    Args:

        state (list) : The fock basis state to use with the class, this should
            be a list of lists, each containing labels for the photons. The
            number of labels in the list will dictate the total photon number
            in the mode.

    """

    __slots__ = ["__s"]

    def __init__(self, state: list[list[int]]) -> None:
        for s in state:
            if not isinstance(s, list):
                raise TypeError("Provided state labels should be lists.")
        self.__s = [sorted(s) for s in state]

    @property
    def n_photons(self) -> int:
        """Returns the number of photons in a State."""
        return sum(len(s) for s in self.__s)

    @property
    def s(self) -> list[list[int]]:
        """Returns list representation of State."""
        return [list(i) for i in self.__s]

    @s.setter
    def s(self, value: Any) -> None:  # noqa: ARG002
        raise AnnotatedStateError(
            "State value should not be modified directly."
        )

    @property
    def n_modes(self) -> int:
        """Returns total number of modes in the State."""
        return len(self.__s)

    @n_modes.setter
    def n_modes(self, value: Any) -> None:  # noqa: ARG002
        raise AnnotatedStateError("Number of modes cannot be modified.")

    def merge(self, merge_state: "AnnotatedState") -> "AnnotatedState":
        """Combine two states, summing the number of photons per mode."""
        if self.n_modes != merge_state.n_modes:
            raise ValueError("Merged states must be the same length.")
        return AnnotatedState(
            [n1 + n2 for n1, n2 in zip(self.__s, merge_state.s, strict=True)]
        )

    def __str__(self) -> str:
        return annotated_state_to_string(self.__s)

    def __repr__(self) -> str:
        return (
            "lightworks.emulator.state.AnnotatedState("
            f"{annotated_state_to_string(self.__s)})"
        )

    def __add__(self, value: "AnnotatedState") -> "AnnotatedState":
        if not isinstance(value, AnnotatedState):
            raise TypeError("Addition only supported between annotated states.")
        return AnnotatedState(self.__s + value.__s)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, AnnotatedState):
            return False
        return self.__s == value.__s

    def __hash__(self) -> int:
        return hash(self.__str__())

    def __len__(self) -> int:
        return self.n_modes

    def __setitem__(self, key: Any, value: Any) -> None:
        raise AnnotatedStateError(
            "AnnotatedState object does not support item assignment."
        )

    def __iter__(self) -> Iterator[list[int]]:
        yield from self.s

    @overload
    def __getitem__(self, indices: int) -> list[int]: ...

    @overload
    def __getitem__(self, indices: slice) -> "AnnotatedState": ...

    def __getitem__(
        self, indices: int | slice
    ) -> Union["AnnotatedState", list[int]]:
        if isinstance(indices, slice):
            return AnnotatedState(self.__s[indices])
        if isinstance(indices, int):
            return self.__s[indices]
        raise TypeError("Subscript should either be int or slice.")



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import NoneType
from typing import Any

import numpy as np
from numpy.typing import NDArray
from scipy.stats import unitary_group


def process_random_seed(seed: Any) -> int | None:
    """
    Takes a provided random seed, validates it is an integer or None, and if
    not will try to convert this. If this isn't possible then an exception will
    be raised.
    """
    if not isinstance(seed, int | NoneType):
        # Try to convert seed to an integer
        try:
            int(seed)
        except Exception as e:
            raise TypeError("Random seed must be an integer.") from e
        # If can convert then check equivalence
        if int(seed) == seed:
            seed = int(seed)
        else:
            raise TypeError("Random seed must be an integer.")
    elif isinstance(seed, bool):
        raise TypeError("Random seed must be an integer.")
    return seed


def random_unitary(
    N: int,  # noqa: N803
    seed: int | None = None,
) -> NDArray[np.complex128]:
    """
    Generate a random NxN unitary matrix. Seed can be used to produce the same
    unitary each time the function is called.

    Args:

        N (int) : The dimension of the random unitary that is to be generated.

        seed (int | None, optional) : Specify a random seed to repeatedly
            produce the same unitary matrix on each function call. Defaults to
            None, which will produce a random matrix on each call.

    Returns:

        np.ndarray : The created random unitary matrix.

    """
    seed = process_random_seed(seed)
    return unitary_group.rvs(N, random_state=seed)


def random_permutation(
    N: int,  # noqa: N803
    seed: int | None = None,
) -> NDArray[np.complex128]:
    """
    Generate a random NxN permutation. Seed can be used to produce the same
    unitary each time the function is called.

    Args:

        N (int) : The dimension of the random permutation that is to be
            generated.

        seed (int | None, optional) : Specify a random seed to repeatedly
            produce the same unitary matrix on each function call. Defaults to
            None, which will produce a random matrix on each call.

    Returns:

        np.ndarray : The created random permutation matrix.

    """
    seed = process_random_seed(seed)
    rng = np.random.default_rng(seed)
    return rng.permutation(np.identity(N, dtype=complex))



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Dedicated unitary component for implementing unitary matrices on a circuit.
"""

import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.utils.param_unitary import ParameterizedUnitary

from .photonic_circuit import PhotonicCircuit
from .photonic_components import UnitaryMatrix


class Unitary(PhotonicCircuit):
    """
    Create a circuit which implements the target provided unitary across all of
    its modes.

    Args:

        unitary (np.ndarray) : The target NxN unitary matrix which is to be
            implemented.

    """

    def __init__(
        self,
        unitary: NDArray[np.complex128] | ParameterizedUnitary,
        label: str = "U",
    ) -> None:
        if not isinstance(unitary, ParameterizedUnitary):
            unitary = np.array(unitary)
        super().__init__(int(unitary.shape[0]))
        self._PhotonicCircuit__circuit_spec = [UnitaryMatrix(0, unitary, label)]



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains a variety of qubit two components, designed for implementing required
qubit processing functionality across pairs of dual-rail encoded qubits in
lightworks.
"""

import numpy as np

from lightworks.sdk.circuit import PhotonicCircuit, Unitary
from lightworks.sdk.utils.permutation import permutation_mat_from_swaps_dict

from .single_qubit_gates import H


class CZ(PhotonicCircuit):
    """
    Post-selected CZ gate that acts across two dual-rail encoded qubits. This
    gate occupies a total of 6 modes, where modes 0 & 5 are used for 0 photon
    heralds, modes 1 & 2 correspond to the 0 & 1 states of the first qubit and
    modes 3 & 4 correspond to the 0 & 1 states of the second qubit. This
    gate requires additional post-selection in which only one photon should be
    measured across each of the pairs of modes which encode a qubit.
    """

    def __init__(self) -> None:
        u_bs = np.array([[-1, 2**0.5], [2**0.5, 1]]) / 3**0.5
        u_a = np.identity(6, dtype=complex)
        for i in range(0, 6, 2):
            u_a[i : i + 2, i : i + 2] = u_bs[:, :]
        u_a[3, :] = -u_a[3, :]
        unitary = Unitary(u_a, label="CZ")
        unitary.herald((0, 0), 0)
        unitary.herald((5, 5), 0)

        super().__init__(4)
        self.add(unitary, 0, group=True, name="CZ")


class CNOT(PhotonicCircuit):
    """
    Post-selected CNOT gate that acts across two dual-rail encoded qubits. This
    gate occupies a total of 6 modes, where modes 0 & 5 are used for 0 photon
    heralds, modes 1 & 2 correspond to the 0 & 1 states of the first qubit and
    modes 3 & 4 correspond to the 0 & 1 states of the second qubit. This gate
    requires additional post-selection in which only one photon should be
    measured across each of the pairs of modes which encode a qubit.

    Args:

        target_qubit (int, optional) : Sets which of the two qubits is used as
            the target qubit for the gate. Should be either 0 or 1.

    """

    def __init__(self, target_qubit: int = 1) -> None:
        if target_qubit not in {0, 1}:
            raise ValueError(
                "target_qubit setting must have a value of either 0 or 1."
            )

        super().__init__(4)

        # Create CNOT from combination of H and CZ
        circ = PhotonicCircuit(4)
        circ.add(H(), 2 * target_qubit)
        circ.add(CZ(), 0)
        circ.add(H(), 2 * target_qubit)

        name = f"CNOT ({1 - target_qubit}, {target_qubit})"
        self.add(circ, 0, group=True, name=name)


class CZ_Heralded(PhotonicCircuit):  # noqa: N801
    """
    Heralded version of the CZ gate which acts across two dual-rail encoded
    qubits, using two NS gates with ancillary photons to herald the success of
    the transformation. This gate occupies 8 modes, where modes 0 & 7 are used
    as 0 photon heralds, modes 1 & 6 are used as 1 photon heralds, mode 2 & 3
    correspond to the 0 & 1 states of the first qubit and modes 4 & 5 correspond
    to the 0 & 1 states of the second qubit. The heralded gate does not require
    any post-selection on the output qubits, other than that they are not lost
    (i.e a total of 4 photons should be measured at the output of the system),
    allowing it to be cascaded with other two qubit gates.
    """

    def __init__(self) -> None:
        u_a = np.identity(8, dtype=complex)

        u_ns = np.array(
            [
                [1 - 2**0.5, 2**-0.25, (3 / (2**0.5) - 2) ** 0.5],
                [2**-0.25, 0.5, 0.5 - 2**-0.5],
                [(3 / (2**0.5) - 2) ** 0.5, 0.5 - 2**-0.5, 2**0.5 - 0.5],
            ]
        )
        u_a[1:4, 1:4] = np.flip(u_ns, axis=(0, 1))[:, :]
        u_a[4:7, 4:7] = u_ns[:, :]
        # Apply pi phase shifts on mode 3
        u_a[:, 3] = -u_a[:, 3]

        # Define beam splitter action
        u_bs = np.identity(8, dtype=complex)
        u_bs[3, 3] = 1 / 2**0.5
        u_bs[4, 4] = 1 / 2**0.5
        u_bs[3, 4] = 1j / 2**0.5
        u_bs[4, 3] = 1j / 2**0.5

        # Define mode reconfiguration so qubits are on central 4 modes
        swaps = {2: 0, 0: 1, 1: 2, 5: 7, 7: 6, 6: 5}
        u_perm1 = permutation_mat_from_swaps_dict(swaps, 8)
        u_perm2 = np.conj(u_perm1.T)

        u_a = u_perm2 @ u_bs @ u_a @ u_bs @ u_perm1

        unitary = Unitary(u_a)
        unitary.herald((0, 0), 0)
        unitary.herald((1, 1), 1)
        unitary.herald((6, 6), 1)
        unitary.herald((7, 7), 0)

        super().__init__(4)
        self.add(unitary, 0, group=True, name="CZ Heralded")


class CNOT_Heralded(PhotonicCircuit):  # noqa: N801
    """
    Heralded version of the CNOT gate which acts across two dual-rail encoded
    qubits, using two NS gates with ancillary photons to herald the success of
    the transformation. This gate occupies 8 modes, where modes 0 & 7 are used
    as 0 photon heralds, modes 1 & 6 are used as 1 photon heralds, mode 2 & 3
    correspond to the 0 & 1 states of the first qubit and modes 4 & 5 correspond
    to the 0 & 1 states of the second qubit. The heralded gate does not require
    any post-selection on the output qubits, other than that they are not lost
    (i.e a total of 4 photons should be measured at the output of the system),
    allowing it to be cascaded with other two qubit gates.

    Args:

        target_qubit (int, optional) : Sets which of the two qubits is used as
            the target qubit for the gate. Should be either 0 or 1.

    """

    def __init__(self, target_qubit: int = 1) -> None:
        if target_qubit not in {0, 1}:
            raise ValueError(
                "target_qubit setting must have a value of either 0 or 1."
            )

        super().__init__(4)

        # Create CNOT from combination of H and CZ
        circ = PhotonicCircuit(4)
        circ.add(H(), 2 * target_qubit)
        circ.add(CZ_Heralded(), 0)
        circ.add(H(), 2 * target_qubit)

        name = f"CNOT Heralded ({1 - target_qubit}, {target_qubit})"
        self.add(circ, 0, group=True, name=name)


class SWAP(PhotonicCircuit):
    """
    Performs a swap between the modes encoding two qubits. To do this, it needs
    to be provided with two tuples, each detailing the two modes used to encode
    the qubit.

    Args:

        qubit_1 (tuple) : A tuple detailing the modes used to encode the 0 & 1
            modes of the first qubit. Should be of the form (m0, m1).

        qubit_2 (tuple) : A tuple detailing the modes used to encode the 0 & 1
            modes of the second qubit. Should be of the form (m0, m1).

    """

    def __init__(
        self, qubit_1: tuple[int, int], qubit_2: tuple[int, int]
    ) -> None:
        # Confirm supplied form is correct
        if len(qubit_1) != 2:
            raise ValueError(
                "qubit_1 value should be a tuple containing two integer mode "
                "numbers."
            )
        if len(qubit_2) != 2:
            raise ValueError(
                "qubit_2 value should be a tuple containing two integer mode "
                "numbers."
            )
        # Extract mode values
        a0, a1 = qubit_1
        b0, b1 = qubit_2
        # Check each mode is an integer
        modes = [a0, a1, b0, b1]
        for m in modes:
            if not isinstance(m, int) or isinstance(m, bool):
                raise TypeError(
                    "One or more mode numbers detected not to be integer."
                )
        n_modes = max(modes) + 1

        # Create circuit with required number of modes
        super().__init__(n_modes)

        # Add required swaps
        self.mode_swaps({a0: b0, b0: a0, a1: b1, b1: a1})



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any


def is_number(value: Any | list[Any]) -> None:
    """
    Function to check if the provided value is float or integer.
    """
    if not isinstance(value, list | tuple):
        value = [value]
    for v in value:
        if not isinstance(v, int | float) or isinstance(v, bool):
            raise TypeError("Distribution values should be a float or integer.")



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from lightworks import (
    ModeMismatchError,
    Parameter,
    PhotonicCircuit,
    PhotonNumberError,
    Simulator,
    State,
    Unitary,
    convert,
    random_unitary,
)
from lightworks.emulator import Backend

BACKEND = Backend("permanent")


class TestSimulator:
    """
    Unit tests to check results returned from simulator in different cases,
    including when fermionic statistics are used.
    """

    def test_hom(self):
        """
        Checks the basic 2 mode hom case and confirms the probability of the
        |0,2> state is 0.5.
        """
        circ = PhotonicCircuit(2)
        circ.bs(0)
        sim = Simulator(circ, State([1, 1]), State([2, 0]))
        results = BACKEND.run(sim)
        assert abs(results.array[0, 0]) ** 2 == pytest.approx(0.5, 1e-8)

    def test_single_photon_case(self):
        """
        Runs a single photon sim and checks the calculated unitary matches the
        target unitary.
        """
        n_modes = 4
        unitary = random_unitary(n_modes)
        unitary_circ = Unitary(unitary)
        states = []
        for i in range(n_modes):
            states.append(State([int(i == j) for j in range(n_modes)]))
        sim = Simulator(unitary_circ, states, states)
        results = BACKEND.run(sim)
        assert (unitary.T.round(8) == results.array.round(8)).all()

    def test_known_result(self):
        """
        Builds a circuit which produces a known result and checks this is found
        at the output.
        """
        # Build circuit
        circuit = PhotonicCircuit(4)
        circuit.bs(1, reflectivity=0.6)
        circuit.mode_swaps({0: 1, 1: 0, 2: 3, 3: 2})
        circuit.bs(0, 3, reflectivity=0.3)
        circuit.bs(0)
        # And check output probability
        sim = Simulator(circuit, State([1, 0, 0, 1]), State([0, 1, 1, 0]))
        results = BACKEND.run(sim)
        assert abs(results.array[0, 0]) ** 2 == pytest.approx(0.5, 1e-8)

    def test_multi_photon_case(self):
        """
        Runs a multi-photon sim and checks the correct value is found for one
        input/output.
        """
        unitary = random_unitary(4, seed=10)
        unitary = Unitary(unitary)
        sim = Simulator(unitary, State([1, 0, 1, 0]), State([0, 2, 0, 0]))
        results = BACKEND.run(sim)
        x = results.array[0, 0]
        assert x == pytest.approx(
            -0.18218877232689196 - 0.266230290128261j, 1e-8
        )

    def test_multi_photon_output_not_specified_case(self):
        """
        Runs a multi-photon sim and checks the correct value is found for one
        input with outputs not specified.
        """
        unitary = random_unitary(4, seed=10)
        unitary = Unitary(unitary)
        sim = Simulator(unitary, State([1, 0, 1, 0]))
        results = BACKEND.run(sim)
        x = results[State([1, 0, 1, 0]), State([0, 2, 0, 0])]
        assert x == pytest.approx(
            -0.18218877232689196 - 0.266230290128261j, 1e-8
        )

    def test_lossy_multi_photon_case(self):
        """
        Runs a lossy multi-photon sim and checks the correct value is found for
        one input/output.
        """
        circ = PhotonicCircuit(4)
        circ.bs(0, loss=convert.db_loss_to_decimal(2))
        circ.ps(1, phi=0.3)
        circ.bs(1, loss=convert.db_loss_to_decimal(2))
        circ.bs(2, loss=convert.db_loss_to_decimal(2))
        circ.ps(1, phi=0.5)
        circ.bs(1, loss=convert.db_loss_to_decimal(2))
        sim = Simulator(circ, State([2, 0, 0, 0]), State([0, 1, 1, 0]))
        results = BACKEND.run(sim)
        x = results.array[0, 0]
        assert x == pytest.approx(
            0.03647550871283556 + 0.01285838825922496j, 1e-8
        )

    def test_circuit_update(self):
        """Used to check circuit updates affect simulator results."""
        unitary = Unitary(random_unitary(4))
        # Create simulator and get initial results
        sim = Simulator(unitary, State([1, 0, 1, 0]), State([0, 2, 0, 0]))
        results = BACKEND.run(sim)
        x = results.array[0, 0]
        # Update circuit adn re-simulate
        unitary.bs(0)
        results = BACKEND.run(sim)
        x2 = results.array[0, 0]
        assert x != x2

    def test_circuit_parameter_update(self):
        """
        Used to check circuit updates through parameter changes affect
        simulator results.
        """
        param = Parameter(0.3)
        circuit = PhotonicCircuit(4)
        circuit.bs(0, reflectivity=param)
        circuit.bs(2, reflectivity=param)
        circuit.bs(1, reflectivity=param)
        # Create simulator and get initial results
        sim = Simulator(circuit, State([1, 0, 1, 0]), State([0, 2, 0, 0]))
        results = BACKEND.run(sim)
        x = results.array[0, 0]
        # Update parameter and re-simulate
        param.set(0.65)
        results = BACKEND.run(sim)
        x2 = results.array[0, 0]
        assert x != x2

    def test_circuit_assignment(self):
        """
        Checks that an incorrect value cannot be assigned to the circuit
        attribute.
        """
        circuit = Unitary(random_unitary(4))
        sim = Simulator(circuit, State([1, 0, 1, 0]))
        with pytest.raises(TypeError):
            sim.circuit = random_unitary(5)

    def test_varied_input_n_raises_error(self):
        """
        Checks that an error is raised if it attempted to use inputs with
        different photon numbers.
        """
        # Create circuit and simulator object
        circuit = Unitary(random_unitary(4))
        # Without output specified
        sim = Simulator(circuit, [State([1, 0, 1, 0]), State([0, 1, 0, 0])])
        with pytest.raises(PhotonNumberError):
            BACKEND.run(sim)
        # With some outputs specified
        sim = Simulator(
            circuit,
            [State([1, 0, 1, 0]), State([0, 1, 0, 0])],
            [State([1, 0, 1, 0]), State([0, 1, 0, 1])],
        )
        with pytest.raises(PhotonNumberError):
            BACKEND.run(sim)

    def test_varied_output_n_raises_error(self):
        """
        Checks that an error is raised if it attempted to use outputs with
        different photon numbers to the input or each other.
        """
        # Create circuit and simulator object
        circuit = Unitary(random_unitary(4))
        # With different number to input
        sim = Simulator(
            circuit,
            [State([1, 0, 1, 0]), State([0, 1, 0, 1])],
            [State([0, 0, 1, 0]), State([0, 1, 0, 0])],
        )
        with pytest.raises(PhotonNumberError):
            BACKEND.run(sim)
        # With different number to each other
        sim = Simulator(
            circuit,
            [State([1, 0, 1, 0]), State([0, 1, 0, 1])],
            [State([1, 0, 1, 0]), State([0, 1, 0, 0])],
        )
        with pytest.raises(PhotonNumberError):
            BACKEND.run(sim)

    def test_incorrect_input_length(self):
        """
        Confirms an error is raised when an input which does not match the
        required circuit modes is supplied.
        """
        # Create circuit and simulator object
        circuit = Unitary(random_unitary(4))
        # Attempt to simulate with input which is too short
        with pytest.raises(ModeMismatchError):
            Simulator(circuit, State([1, 0, 1]))
        # And then which is too long
        with pytest.raises(ModeMismatchError):
            Simulator(circuit, State([1, 0, 1, 0, 1]))

    def test_incorrect_input_length_herald(self):
        """
        Confirms an error is raised when an input which does not match the
        required circuit modes is supplied, when heralding is used in the
        original circuit.
        """
        # Create circuit and simulator object
        circuit = Unitary(random_unitary(4))
        circuit.herald((0, 2), 1)
        # Attempt to simulate with input which is too short
        with pytest.raises(ModeMismatchError):
            Simulator(circuit, State([1, 0]))
        # And then which is too long
        with pytest.raises(ModeMismatchError):
            Simulator(circuit, State([1, 0, 1, 0]))

    def test_incorrect_input_length_herald_grouped(self):
        """
        Confirms an error is raised when an input which does not match the
        required circuit modes is supplied, when heralding is used in a
        sub-circuit of the original circuit.
        """
        # Create circuit and simulator object
        circuit = Unitary(random_unitary(4))
        sub_circuit = Unitary(random_unitary(4))
        sub_circuit.herald((0, 2), 1)
        circuit.add(sub_circuit, 1)
        # Attempt to simulate with input which is too short
        with pytest.raises(ModeMismatchError):
            Simulator(circuit, State([1, 0, 1]))
        # And then which is too long
        with pytest.raises(ModeMismatchError):
            Simulator(circuit, State([1, 0, 1, 0, 0]))

    @pytest.mark.parametrize("n_output", [0, 1, 2])
    def test_herald_not_herald_equivalance(self, n_output):
        """
        Checks that the results from the Simulator are equivalent when using a
        heralded circuit and getting the same outputs from a non-heralded
        circuit.
        """
        # Create heralded and versions of same circuit
        unitary = random_unitary(6)
        circuit = Unitary(unitary)
        circuit_herald = Unitary(unitary)
        circuit_herald.herald((2, 2), 0)
        circuit_herald.herald((1, 3), (1, n_output))
        # Simulate both with equivalent inputs
        sim = Simulator(circuit, State([0, 1, 0, 1, 1, 0]))
        results = BACKEND.run(sim)
        sim_h = Simulator(circuit_herald, State([0, 1, 1, 0]))
        results_h = BACKEND.run(sim_h)
        # Then check equivalence of results for all outputs
        for output in results_h.outputs:
            full_state = output[0:2] + State([0, n_output]) + output[2:]
            assert (
                pytest.approx(results_h[State([0, 1, 1, 0]), output])
                == results[State([0, 1, 0, 1, 1, 0]), full_state]
            )

    @pytest.mark.parametrize("n_output", [0, 1, 2])
    def test_herald_not_herald_equivalance_lossy(self, n_output):
        """
        Checks that the results from the Simulator are equivalent when using a
        heralded circuit and getting the same outputs from a non-heralded
        circuit, while including loss modes.
        """
        # Create heralded and versions of same circuit
        unitary = random_unitary(6)
        circuit = Unitary(unitary)
        for i in range(6):
            circuit.loss(i, (i + 1) / 10)
        circuit_herald = Unitary(unitary)
        for i in range(6):
            circuit_herald.loss(i, (i + 1) / 10)
        circuit_herald.herald((2, 2), 0)
        circuit_herald.herald((1, 3), (1, n_output))
        # Simulate both with equivalent inputs
        sim = Simulator(circuit, State([0, 1, 0, 1, 1, 0]))
        results = BACKEND.run(sim)
        sim_h = Simulator(circuit_herald, State([0, 1, 1, 0]))
        results_h = BACKEND.run(sim_h)
        # Then check equivalence of results for all outputs
        for output in results_h.outputs:
            full_state = output[0:2] + State([0, n_output]) + output[2:]
            assert (
                pytest.approx(results_h[State([0, 1, 1, 0]), output])
                == results[State([0, 1, 0, 1, 1, 0]), full_state]
            )

    @pytest.mark.parametrize("n_output", [0, 1, 2])
    def test_herald_not_herald_equivalance_grouped(self, n_output):
        """
        Checks that the results from the Simulator are equivalent when using a
        heralded circuit and getting the same outputs from a non-heralded
        circuit, while the heralded circuit contains heralds within a grouped
        sub-circuit.
        """
        # Create heralded and versions of same circuit
        unitary = random_unitary(6)
        circuit = Unitary(unitary)
        for i in range(6):
            circuit.loss(i, (i + 1) / 10)
        sub_circuit = Unitary(unitary)
        for i in range(6):
            sub_circuit.loss(i, (i + 1) / 10)
        sub_circuit.herald((2, 2), 0)
        sub_circuit.herald((1, 3), (1, n_output))
        circuit_herald = PhotonicCircuit(4)
        circuit_herald.add(sub_circuit)
        # Simulate both with equivalent inputs
        sim = Simulator(circuit, State([0, 1, 0, 1, 1, 0]))
        results = BACKEND.run(sim)
        sim_h = Simulator(circuit_herald, State([0, 1, 1, 0]))
        results_h = BACKEND.run(sim_h)
        # Then check equivalence of results for all outputs
        for output in results_h.outputs:
            full_state = output[0:2] + State([0, n_output]) + output[2:]
            assert (
                pytest.approx(results_h[State([0, 1, 1, 0]), output])
                == results[State([0, 1, 0, 1, 1, 0]), full_state]
            )

    def test_output_heralds_too_large(self):
        """
        Confirms a PhotonNumberError is raised when the number of output heralds
        is larger than the number of photons input into the system.
        """
        circuit = Unitary(random_unitary(4))
        circuit.herald(3, (0, 2))
        sim = Simulator(circuit, State([1, 0, 0]))
        with pytest.raises(PhotonNumberError):
            sim._generate_task()



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from lightworks import State
from lightworks.emulator import AnnotatedStateError
from lightworks.emulator.state import AnnotatedState
from lightworks.emulator.utils.state import annotated_state_to_string


class TestAnnotatedState:
    """Unit tests for AnnotatedState object."""

    def test_get_mode_value(self):
        """Confirms that mode value retrieval works as expected."""
        s = AnnotatedState([[], [0], [1], []])
        assert s[1] == [0]

    def test_state_equality(self):
        """Checks that state equality comparison works as expected."""
        assert AnnotatedState([[], [0], [1], []]) == AnnotatedState(
            [[], [0], [1], []]
        )
        assert AnnotatedState([[], [0], [1], []]) != AnnotatedState(
            [[], [1], [0], []]
        )

    def test_state_equality_diff_types(self):
        """
        Checks that state equality returns False when compared with a
        non-AnnotatedState object.
        """
        s = AnnotatedState([[0], [1], [2]])
        assert s != [[0], [1], [2]]

    def test_state_equality_shuffled(self):
        """
        Checks that state equality comparison works as expected in the case
        that identical labels are provide in a mode but in a different order.
        """
        assert AnnotatedState([[], [0], [1, 2], []]) == AnnotatedState(
            [[], [0], [2, 1], []]
        )

    def test_state_addition(self):
        """Checks that state addition behaviour is as expected."""
        s = AnnotatedState([[0], [2, 3], [1]]) + AnnotatedState(
            [[], [0, 3], [1]]
        )
        assert s == AnnotatedState([[0], [2, 3], [1], [], [0, 3], [1]])
        s = AnnotatedState([[0], [2, 3], [1]]) + AnnotatedState([])
        assert s == AnnotatedState([[0], [2, 3], [1]])

    def test_merge(self):
        """Checks that state merge works correctly."""
        s = AnnotatedState([[0], [2, 3], [1]]).merge(
            AnnotatedState([[], [0], [1]])
        )
        assert s == AnnotatedState([[0], [2, 3, 0], [1, 1]])

    def test_modification_behavior(self):
        """
        Checks that the correct error is raised if we try to modify the State
        value. Also tests what happens when state s attribute is modified.
        """
        s = AnnotatedState([[0, 1], [2]])
        with pytest.raises(AnnotatedStateError):
            s.s = [[0]]
        s.s[0] = [2]
        assert s == AnnotatedState([[0, 1], [2]])

    def test_mode_length(self):
        """Check the calculated mode number attribute is set correctly."""
        assert len(AnnotatedState([[], [0], [1, 2], [], [0]])) == 5
        assert len(AnnotatedState([[], [], [], []])) == 4
        assert len(AnnotatedState([])) == 0

    def test_photon_number(self):
        """Checks calculated photon number value is correct."""
        assert AnnotatedState([[], [0], [1, 2], [], [0]]).n_photons == 4
        assert AnnotatedState([[], [], [], []]).n_photons == 0
        assert AnnotatedState([]).n_photons == 0

    @pytest.mark.parametrize(
        "value", [[0, 1, 2], ["0", "1", "2"], [[0], 1], [0], [None]]
    )
    def test_req_list_of_lists(self, value):
        """
        Checks that a exception is raised in Annotated state isn't provided a
        list of lists as the input.
        """
        with pytest.raises(TypeError):
            AnnotatedState(value)

    def test_n_modes_modification(self):
        """
        Tests that an exception is raised when a user attempts to modify an
        AnnotatedState.
        """
        s = AnnotatedState([[0], [1], [2]])
        with pytest.raises(AnnotatedStateError):
            s.n_modes = 2

    def test_merge_not_same_length(self):
        """
        Confirms that merge method raises an exception when AnnotatedStates are
        not of the same length.
        """
        s = AnnotatedState([[0], [1], [2]])
        s2 = AnnotatedState([[0], [1], [2], [3]])
        with pytest.raises(ValueError):
            s.merge(s2)
        with pytest.raises(ValueError):
            s2.merge(s)

    @pytest.mark.parametrize("value", [(0, 1, 2), [0, 1, 2], State([0, 1, 2])])
    def test_add_not_same_type(self, value):
        """
        Checks that an exception is raised when a non-AnnotatedState is
        attempted to be added to an AnnotatedState
        """
        s = AnnotatedState([[0], [1], [2]])
        with pytest.raises(TypeError):
            s + value

    def test_str_returns_correct_value(self):
        """
        Checks that string function returns correct value for AnnotatedState.
        """
        s = [[0], [1], [2]]
        assert str(AnnotatedState(s)) == annotated_state_to_string(s)

    def test_repr_returns_correct_value(self):
        """
        Checks that repr function contains correct value for AnnotatedState.
        """
        s = [[0], [1], [2]]
        assert annotated_state_to_string(s) in str(AnnotatedState(s))

    def test_set_item_blocked(self):
        """
        Confirms that set item functionality is blocked for the AnnotatedState.
        """
        s = AnnotatedState([[0], [1], [2]])
        with pytest.raises(AnnotatedStateError):
            s[0] = [1]

    def test_get_item_slicing(self):
        """
        Checks that slicing can be used to get a part of an AnnotatedState.
        """
        s = AnnotatedState([[0], [1], [2], [3]])
        assert s[0:2] == AnnotatedState([[0], [1]])

    @pytest.mark.parametrize("value", [[0, 1], None, "1"])
    def test_get_item_invalid_value(self, value):
        """
        Confirms error raised when an invalid value is provided to get item.
        """
        s = AnnotatedState([[0], [1], [2], [3]])
        with pytest.raises(TypeError):
            s[value]



# Copyright 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import numpy as np
import sympy as sp
from numpy.typing import NDArray

from lightworks.sdk.circuit.parameters import Parameter


class ParameterizedUnitary:
    """
    Adds support for parametrisation of unitaries for delayed computation with
    sympy.

    Args:

        unitary (sp.Matrix) : The sympy matrix which represents the unitary to
            be implemented.

        params (dict) : A dictionary where the keys should be the names of the
            symbols used in the sympy array and the values are Lightworks
            parameters.

    """

    def __init__(
        self, unitary: sp.Matrix, params: dict[str, Parameter[Any]]
    ) -> None:
        if not isinstance(unitary, sp.Matrix):
            raise TypeError("Supplied unitary must be a sympy matrix.")
        self._unitary = unitary
        self._params = params

    def __str__(self) -> str:
        return str(self._unitary)

    def __repr__(self) -> str:
        return (
            f"ParameterizedUnitary(matrix={self._unitary}, "
            f"params={self._params})"
        )

    @property
    def unitary(self) -> NDArray[np.complex128]:
        """Returns the current value of the unitary matrix."""
        vals = {
            k: p.get() if isinstance(p, Parameter) else p
            for k, p in self._params.items()
        }
        return np.array(self._unitary.evalf(subs=vals), dtype=complex)

    @property
    def shape(self) -> tuple[int, int]:
        """Returns current shape of the unitary matrix."""
        return self._unitary.shape



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.state import State

from .experiments import ProcessTomographyExperiment, ProcessTomographyList
from .mappings import INPUT_MAPPING, MEASUREMENT_MAPPING, RHO_MAPPING
from .tomography import _Tomography
from .utils import (
    TomographyDataError,
    _calculate_density_matrix,
    _check_target_process,
    _combine_all,
    _get_required_tomo_measurements,
    _get_tomo_measurements,
    state_fidelity,
)


class _ProcessTomography(_Tomography):
    """
    Process tomography base class, implements some of the common methods
    required across different approaches.
    """

    _tomo_inputs: tuple[str, ...] = ("Z+", "Z-", "X+", "Y+")

    def get_experiments(self) -> ProcessTomographyList:
        """
        Generates all required tomography experiments for performing a process
        tomography algorithm.
        """
        inputs = self._full_input_basis()
        req_measurements, _ = _get_required_tomo_measurements(self.n_qubits)
        # Determine required input states and circuits
        experiments = ProcessTomographyList()
        for in_basis in inputs:
            for meas_basis in req_measurements:
                circ, state = self._create_circuit_and_input(
                    in_basis, meas_basis
                )
                experiments.append(
                    ProcessTomographyExperiment(
                        circuit=circ,
                        input_state=state,
                        input_basis=in_basis,
                        measurement_basis=meas_basis,
                    )
                )

        return experiments

    def process_state_fidelities(
        self,
        data: list[dict[State, int]] | dict[tuple[str, str], dict[State, int]],
        target_process: NDArray[np.complex128],
    ) -> dict[str, float]:
        """
        Calculates the density matrix of the state created for each input to the
        system and finds the fidelity relative to the expected matrix.

        Args:

            data (list | dict) : The collected measurement data. If a list then
                this should match the order the experiments were provided, and
                if a dictionary, then each key should be tuple of the input and
                measurement basis.

            target_process (np.ndarray) : The unitary matrix corresponding to
                the target process. The dimension of this should be 2^n_qubits.

        Returns:

            float : The calculated fidelity.

        """
        target_process = _check_target_process(target_process, self.n_qubits)
        rhos = self.process_density_matrices(data)
        rho_exp = self.get_expected_density_matrices(target_process)
        return {i: state_fidelity(r, rho_exp[i]) for i, r in rhos.items()}

    def process_density_matrices(
        self,
        data: list[dict[State, int]] | dict[tuple[str, str], dict[State, int]],
    ) -> dict[str, NDArray[np.complex128]]:
        """
        Calculate and returns the density matrix of the state created for each
        input to the system.

        Args:

            data (list | dict) : The collected measurement data. If a list then
                this should match the order the experiments were provided, and
                if a dictionary, then each key should be tuple of the input and
                measurement basis.

        Returns:

            dict : A dictionary of the calculated density matrix for each input
                to the system.

        """
        # Run all required tomography experiments
        results = self._convert_tomography_data(data)
        # Sorted results per input
        remapped_results: dict[str, dict[str, dict[State, int]]] = {
            k[0]: {} for k in results
        }
        for (k1, k2), r in results.items():
            remapped_results[k1][k2] = r
        # Calculate density matrices
        return {
            i: _calculate_density_matrix(remapped_results[i], self.n_qubits)
            for i in self._full_input_basis()
        }

    def get_expected_density_matrices(
        self, target_process: NDArray[np.complex128]
    ) -> dict[str, NDArray[np.complex128]]:
        """
        Calculates the expected output density matrices for each input when a
        target process is applied.

        Args:

            target_process (np.ndarray) : The unitary matrix corresponding to
                the target process. The dimension of this should be 2^n_qubits.

        Returns:

            dict : The calculated density matrices for each input.

        """
        target_process = _check_target_process(target_process, self.n_qubits)
        input_rhos = _combine_all(
            {k: v for k, v in RHO_MAPPING.items() if k in self._tomo_inputs},
            self.n_qubits,
        )
        return {
            i: target_process @ input_rhos[i] @ np.conj(target_process.T)
            for i in self._full_input_basis()
        }

    def _full_input_basis(self) -> list[str]:
        """
        Calculates and returns the full input basis used in a tomography
        experiment.
        """
        return _combine_all(list(self._tomo_inputs), self.n_qubits)

    def _create_circuit_and_input(
        self, input_op: str, output_op: str
    ) -> tuple[PhotonicCircuit, State]:
        """
        Creates the required circuit and input state to achieve a provided input
        and output operation.
        """
        in_state = State([])
        circ = PhotonicCircuit(self.base_circuit.input_modes)
        # Input operation
        for i, op in enumerate(input_op.split(",")):
            in_state += INPUT_MAPPING[op][0]
            circ.add(INPUT_MAPPING[op][1], 2 * i)
        # Add base circuit
        circ.add(self.base_circuit)
        # Measurement operation
        for i, op in enumerate(output_op.split(",")):
            circ.add(MEASUREMENT_MAPPING[op], 2 * i)
        return circ, in_state

    def _convert_tomography_data(
        self,
        results: list[dict[State, int]]
        | dict[tuple[str, str], dict[State, int]],
    ) -> dict[tuple[str, str], dict[State, int]]:
        """
        Processes the provided tomography data, validating all expected
        measurements are present and converting it into a full dataset including
        previously removed redundant measurements (in the 'I' basis).
        """
        # Re-generate all tomography data
        inputs = _combine_all(list(self._tomo_inputs), self.n_qubits)
        req_measurements, result_mapping = _get_required_tomo_measurements(
            self.n_qubits
        )
        if not isinstance(results, dict):
            if len(results) != len(inputs) * len(req_measurements):
                msg = (
                    f"Number of results ({len(results)}) did not match the "
                    f"expected number ({len(inputs) * len(req_measurements)}) "
                    "for the target tomography algorithm."
                )
                raise TomographyDataError(msg)
            # Sort results into each input/measurement combination
            num_per_in = len(req_measurements)
            sorted_results = {
                in_state: dict(
                    zip(
                        req_measurements,
                        results[num_per_in * i : num_per_in * (i + 1)],
                        strict=True,
                    )
                )
                for i, in_state in enumerate(inputs)
            }
        else:
            sorted_results = {}
            missing = []
            for in_state in inputs:
                sorted_results[in_state] = {}
                for meas in req_measurements:
                    if (in_state, meas) in results:
                        sorted_results[in_state][meas] = results[in_state, meas]
                    else:
                        missing.append((in_state, meas))
            if missing:
                msg = (
                    "One or more expected keys were detected to be missing "
                    f"from the results dictionary. Missing keys were {missing}."
                )
                raise TomographyDataError(msg)
        # Expand results to include all of the required measurements
        full_results = {}
        for in_state, res in sorted_results.items():
            for meas in _get_tomo_measurements(self.n_qubits):
                full_results[in_state, meas] = res[result_mapping[meas]]
        return full_results



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import drawsvg as draw
import numpy as np

from lightworks.sdk.utils.exceptions import DisplayError


class DrawSVGComponents:
    """
    Manages addition of all components to the Drawing.
    """

    def __init__(self, drawing: draw.Drawing, wg_width: float) -> None:
        self.d = drawing
        self.wg_width = wg_width

    def add(self, draw_spec: list[tuple[str, tuple[Any, ...]]]) -> None:
        """
        Adds components to the provided drawing using the draw spec.
        """
        # Loop over each element in the drawing spec and add
        for c, data in draw_spec:
            if c == "wg":
                self._draw_wg(*data)
            elif c == "ps":
                self._draw_ps(*data)
            elif c == "bs":
                self._draw_bs(*data)
            elif c == "text":
                self._draw_text(*data)
            elif c == "lc":
                self._draw_loss(*data)
            elif c == "mode_swaps":
                self._draw_mode_swaps(*data)
            elif c == "unitary":
                self._draw_unitary(*data)
            elif c == "group":
                self._draw_grouped_circuit(*data)
            elif c == "herald":
                self._draw_herald(*data)
            else:
                raise DisplayError("Element in draw spec not recognised.")

    def _draw_wg(self, x: float, y: float, length: float) -> None:
        r = draw.Rectangle(
            x, y - self.wg_width / 2, length, self.wg_width, fill="black"
        )
        self.d.append(r)

    def _draw_ps(self, x: float, y: float, size: float) -> None:
        r = draw.Rectangle(
            x,
            y - size / 2,
            size,
            size,
            fill="#e8532b",
            stroke="black",
            rx=5,
            ry=5,
        )
        self.d.append(r)

    def _draw_bs(
        self, x: float, y: float, size_x: float, size_y: float, offset_y: float
    ) -> None:
        r = draw.Rectangle(
            x,
            y - offset_y,
            size_x,
            size_y,
            fill="#3e368d",
            stroke="black",
            rx=5,
            ry=5,
        )
        self.d.append(r)

    def _draw_unitary(
        self, x: float, y: float, size_x: float, size_y: float, offset_y: float
    ) -> None:
        r = draw.Rectangle(
            x,
            y - offset_y,
            size_x,
            size_y,
            fill="#1a0f36",
            stroke="black",
            rx=5,
            ry=5,
        )
        self.d.append(r)

    def _draw_loss(self, x: float, y: float, size: float) -> None:
        r = draw.Rectangle(
            x, y - size / 2, size, size, fill="grey", stroke="black", rx=5, ry=5
        )
        self.d.append(r)

    def _draw_text(
        self,
        text: str,
        x: float,
        y: float,
        rotation: float,
        size: float,
        colour: str,
        alignment: str,
    ) -> None:
        if alignment == "centred":
            ta = "middle"
            db = "middle"
        elif alignment == "left":
            ta = "start"
            db = "middle"
        elif alignment == "right":
            ta = "end"
            db = "middle"
        else:
            raise DisplayError("Alignment value not recognised.")
        t = draw.Text(
            text,
            size,
            x,
            y,
            fill=colour,
            text_anchor=ta,
            dominant_baseline=db,
            transform=f"rotate({rotation}, {x}, {y})",
        )
        self.d.append(t)

    def _draw_mode_swaps(
        self, x: float, ys: list[tuple[float, float]], size_x: float
    ) -> None:
        for y0, y1 in ys:
            w = self.wg_width / 2
            m = np.arctan(abs(y1 - y0) / size_x)
            if y0 < y1:
                dx1 = w * m
                dx2 = 0
            else:
                dx1 = 0
                dx2 = w * m

            points = [
                x + dx1,
                y0 - w,
                x,
                y0 - w,
                x,
                y0 + w,
                x + dx2,
                y0 + w,
                x + size_x - dx1,
                y1 + w,
                x + size_x,
                y1 + w,
                x + size_x,
                y1 - w,
                x + size_x - dx2,
                y1 - w,
            ]
            poly = draw.Lines(*points, fill="black", close=True)
            self.d.append(poly)

    def _draw_grouped_circuit(
        self, x: float, y: float, size_x: float, size_y: float, offset_y: float
    ) -> None:
        r = draw.Rectangle(
            x,
            y - offset_y,
            size_x,
            size_y,
            fill="#1a0f36",
            stroke="black",
            rx=5,
            ry=5,
        )
        self.d.append(r)

    def _draw_herald(self, x: float, y: float, size: float) -> None:
        c = draw.Circle(x, y, size, fill="#3e368d", stroke="black")
        self.d.append(c)



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Lightworks
==========

Lightworks is a high-level interface for photonic quantum computing,
enabling a range of transformations to be implemented and executed on both
remote photonic hardware and with local simulation tools.

Key objects:

    Circuit : Provides a tool for building linear optic circuits across a
        number of modes. Circuit also supports addition for combination of
        sub-circuits and can be used with built-in Parameter objects to allow
        for adjustment of the circuit configuration after creation. Circuit
        also has a display method so the created circuit can be viewed.

    State : Represents the photonic fock states which are input and output from
        the system. State objects are hashable, and so can be used as keys in
        dictionaries, and support both addition and merging to combine states
        together.

    emulator : Provides a set of local simulation tools for the testing and
        verification of outputs from a given problem. There is a number of
        different objects for simulation of the system, which provide various
        capabilities and outputs.

    qubit : Module for implementing the qubit paradigm of quantum computing on
        photonic linear optic systems.

    tomography : Module for performing quantum tomography on linear optic
        gate-based circuits.

"""

import contextlib

from . import emulator, interferometers, qubit, tomography
from .__settings import settings
from .__version import __version__  # noqa: F401
from .sdk.circuit import Parameter, ParameterDict, PhotonicCircuit, Unitary
from .sdk.state import State
from .sdk.tasks import Analyzer, Batch, Sampler, Simulator
from .sdk.utils import (
    PostSelection,
    PostSelectionFunction,
    convert,
    random_permutation,
    random_unitary,
)
from .sdk.utils.exceptions import *  # noqa: F403
from .sdk.visualisation import display

# If installed then also import the remote module
with contextlib.suppress(ModuleNotFoundError):
    import lightworks_remote as remote  # noqa: F401

__all__ = [
    "Analyzer",
    "Batch",
    "Parameter",
    "ParameterDict",
    "PhotonicCircuit",
    "PostSelection",
    "PostSelectionFunction",
    "Sampler",
    "Simulator",
    "State",
    "Unitary",
    "convert",
    "display",
    "emulator",
    "interferometers",
    "qubit",
    "random_permutation",
    "random_unitary",
    "settings",
    "tomography",
]



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains a number of functions for converting between quantities in the
emulator.
"""

from math import log10

from lightworks.sdk.state import State


class convert:  # noqa: N801
    """
    Contains a range of functions for converting/mapping between quantities.
    """

    @staticmethod
    def db_loss_to_decimal(loss: float) -> float:
        """
        Function to convert from a given dB loss into the equivalent loss value
        in decimal form. Note this function does not support conversion of gain
        values.

        Args:

            loss (float) : The loss value in decibels.

        Returns:

            float : The calculated loss as a decimal.

        """
        # Standardize loss format
        loss = -abs(loss)
        return 1 - 10 ** (loss / 10)

    @staticmethod
    def decimal_to_db_loss(loss: float) -> float:
        """
        Function to convert from a decimal into dB loss. This dB loss will be
        returned as a positive value.

        Args:

            loss (float) : The loss value as a decimal, this should be in the
                range [0,1).

        Returns:

            float : The calculated dB loss. This is returned as a positive
                value.

        Raises:

            ValueError: Raised in cases where transmission is not in range
                [0,1).

        """
        if loss < 0 or loss >= 1:
            raise ValueError("Transmission value should be in range [0,1).")
        return abs(10 * log10(1 - loss))

    @staticmethod
    def qubit_to_dual_rail(state: State | list[int] | str) -> State:
        """
        Converts from a qubit encoding into a dual-rail encoded state between
        modes.

        Args:

            state (State) : The qubit state to convert.

        Returns:

            State : The dual-rail encoded Fock state.

        Raises:

            ValueError: Raised when values in the original state aren't either
                0 or 1.

        """
        new_state = []
        for s in state:
            if s in {"0", "1"}:  # Support string values
                s = int(s)  # noqa: PLW2901
            if s not in {0, 1}:
                raise ValueError(
                    "Elements of a qubit state can only take integer values 0 "
                    "or 1."
                )
            new_state += [1, 0] if not s else [0, 1]
        return State(new_state)

    @staticmethod
    def dual_rail_to_qubit(
        state: State | list[int], allow_invalid: bool = False
    ) -> State:
        """
        Converts from a dual-rail encoded Fock state into the qubit encoded
        equivalent.

        Args:

            state (State) : The dual-rail state to convert. This state should
                contain a single photon between pairs of adjacent modes.

            allow_invalid (bool) : Controls whether or not invalid values are
                supported for a qubit. In these cases, the numerical value will
                be replaced by an X.

        Returns:

            State : The calculated qubit state.

        Raises:

            ValueError: Raised when an invalid state is provided for conversion.

        """
        new_state: list[int | str] = []
        if len(state) % 2 != 0:
            raise ValueError(
                "Dual-rail encoded state should have an even number of modes."
            )
        list_state = list(state)
        for i in range(len(state) // 2):
            sub_s = list_state[2 * i : 2 * i + 2]
            if sub_s not in ([1, 0], [0, 1]):
                if not allow_invalid:
                    raise ValueError(
                        "Invalid entry found in state. State should have a "
                        "single photon between each pair of dual-rail encoded "
                        "modes. Alternatively, set allow_invalid = True."
                    )
                new_state.append("X")
            else:
                new_state.append(sub_s[1])
        return State(new_state)  # type: ignore[arg-type]

    @staticmethod
    def threshold_mapping(state: State, invert: bool = False) -> State:
        """
        Apply a threshold mapping to a State, in which any values above 1 will
        be reduced to 1.

        Args:

            state (State) : The state to be converted.

            invert (bool, optional) : Select whether to invert the threshold
                mapping. This will swap the 0s and 1s of the produced states.

        Returns:

            State : The threshold mapped state.

        """
        return State([abs(min(s, 1) - invert) for s in state])

    @staticmethod
    def parity_mapping(state: State, invert: bool = False) -> State:
        """
        Apply a parity mapping to a State, in which the state values are
        converted to either 0 or 1 depending on if the value is odd or even.

        Args:

            state (State) : The state to be converted.

            invert (bool, optional) : Select whether to invert the parity
                mapping. This will swap between even->0 & odd->1 and even->1 &
                odd->0.

        Returns:

            State : The parity mapped state.

        """
        return State([abs((s % 2) - invert) for s in state])



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from lightworks.sdk.circuit import PhotonicCircuit

from .decomposition import reck_decomposition
from .error_model import ErrorModel


class Reck:
    """
    Maps a provided PhotonicCircuit to a Reck interferometer, implementing the
    configured error model across all components.

    Args:

        error_model (ErrorModel | None) : An error model object which implements
            variation of the interferometer parameters. If not specified then
            an ideal system will be assumed.

    """

    def __init__(self, error_model: ErrorModel | None = None) -> None:
        # If no error model provided then set to default
        if error_model is None:
            error_model = ErrorModel()
        self.error_model = error_model

    @property
    def error_model(self) -> ErrorModel:
        """Returns currently used error model for the system."""
        return self.__error_model

    @error_model.setter
    def error_model(self, value: ErrorModel) -> None:
        if not isinstance(value, ErrorModel):
            raise TypeError("error_model should be an ErrorModel object.")
        self.__error_model = value

    def map(
        self, circuit: PhotonicCircuit, seed: int | None = None
    ) -> PhotonicCircuit:
        """
        Maps a provided circuit onto the interferometer.

        Args:

            circuit (PhotonicCircuit) : The circuit to map onto the reck
                interferometer.

            seed (int | None) : Set a random seed to produce identical results
                when implementing a probabilistic error model.

        Returns:

            PhotonicCircuit : The created Reck interferometer which implements
                the programmed transformation.

        """
        # Reset error model seed
        self.error_model._set_random_seed(seed)
        # Invert unitary so reck layout starts with fewest elements on mode 0
        unitary = np.flip(circuit.U, axis=(0, 1))
        phase_map, end_phases = reck_decomposition(unitary)
        phase_map = {
            k: (v + self.error_model.get_phase_offset()) % (2 * np.pi)
            for k, v in phase_map.items()
        }
        end_phases = [
            (p + self.error_model.get_phase_offset()) % (2 * np.pi)
            for p in end_phases
        ]

        # Build circuit with required mode number
        n_modes = circuit.n_modes
        mapped_circuit = PhotonicCircuit(n_modes)
        for i in range(n_modes - 1):
            for j in range(0, n_modes - 1 - i, 1):
                # Find coordinates + mode from i & j values
                coord = f"{j + 2 * i}_{j}"
                mode = n_modes - j - 2
                mapped_circuit.barrier([mode, mode + 1])
                mapped_circuit.ps(mode + 1, phase_map["ps_" + coord])
                mapped_circuit.bs(
                    mode, reflectivity=self.error_model.get_bs_reflectivity()
                )
                mapped_circuit.ps(mode, phase_map["bs_" + coord])
                mapped_circuit.bs(
                    mode,
                    reflectivity=self.error_model.get_bs_reflectivity(),
                    loss=self.error_model.get_loss(),
                )
        mapped_circuit.barrier()
        # Apply residual phases at the end
        for i in range(n_modes):
            mapped_circuit.ps(n_modes - i - 1, end_phases[i])

        # Add any heralds from the original circuit
        heralds = circuit.heralds
        for m1, m2 in zip(heralds.input, heralds.output, strict=True):
            mapped_circuit.herald(
                (m1, m2), (heralds.input[m1], heralds.output[m2])
            )

        return mapped_circuit



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any

import matplotlib.figure
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches
from multimethod import multimethod

from lightworks.sdk.circuit.photonic_components import (
    Barrier,
    BeamSplitter,
    Group,
    HeraldData,
    Loss,
    ModeSwaps,
    PhaseShifter,
    UnitaryMatrix,
)
from lightworks.sdk.utils.exceptions import DisplayError

from .display_utils import process_parameter_value

if TYPE_CHECKING:
    from lightworks.sdk.circuit import PhotonicCircuit


class DrawCircuitMPL:
    """
    DrawCircuit

    This class can be used to Display a circuit in the quantum emulator as a
    figure in matplotlib.

    Args:

        circuit (PhotonicCircuit) : The circuit which is to be displayed.

        display_loss (bool, optional) : Choose whether to display loss
            components in the figure, defaults to False.

        mode_label (list|None, optional) : Optionally provided a list of mode
            labels which will be used to name the mode something other than
            numerical values. Can be set to None to use default values.

        show_parameter_values (bool, optional) : Shows the values of parameters
            instead of the associated labels if specified.

    """

    def __init__(
        self,
        circuit: "PhotonicCircuit",
        display_loss: bool = False,
        mode_labels: list[str] | None = None,
        show_parameter_values: bool = False,
    ) -> None:
        self.circuit = circuit
        self.display_loss = display_loss
        self.mode_labels = mode_labels
        self.show_parameter_values = show_parameter_values
        self.n_modes = self.circuit.n_modes
        self.herald_modes = self.circuit._internal_modes

    def draw(self) -> tuple[matplotlib.figure.Figure, plt.Axes]:
        """
        Creates matplotlib figure using the provided circuit spec.
        """
        # Set a waveguide width and get mode number
        self.wg_width = 0.05
        # Adjust size of figure according to circuit with min size 4 and max 40
        s = min(len(self.circuit._get_circuit_spec()), 40)
        s = max(s, 4)
        # Create fig and set aspect to equal
        self.fig, self.ax = plt.subplots(figsize=(s, s), dpi=200)
        self.ax.set_aspect("equal")
        # Manually adjust figure height
        h = max(self.n_modes, 4)
        self.fig.set_figheight(h)
        dy = 1.0
        dy_smaller = 0.6
        self.y_locations = []
        # Set mode y locations
        yloc = 0.0
        for i in range(self.n_modes):
            self.y_locations.append(yloc)
            if i + 1 in self.herald_modes or i in self.herald_modes:
                yloc += dy_smaller
            else:
                yloc += dy
        # Set a starting length and add a waveguide for each mode
        init_length = 0.5
        if False:
            self._add_wg(0, i - self.wg_width / 2, init_length)
        # Create a list to store the positions of each mode
        self.x_locations = [init_length] * self.n_modes
        # Add extra waveguides when using heralds
        if self.circuit._external_heralds.input:
            for i in range(self.n_modes):
                if i not in self.herald_modes:
                    self._add_wg(self.x_locations[i], self.y_locations[i], 0.5)
                    self.x_locations[i] += 0.5
        # Loop over circuit spec and add each component
        for spec in self.circuit._get_circuit_spec():
            self._add(spec)

        # Add any final lengths as required
        final_loc = max(self.x_locations)
        # Extend final waveguide if herald included
        if self.circuit._external_heralds.output:
            final_loc += 0.5
        for i, loc in enumerate(self.x_locations):
            if loc < final_loc and i not in self.herald_modes:
                length = final_loc - loc
                self._add_wg(loc, self.y_locations[i], length)
                self.x_locations[i] += length

        # Add heralding display
        self._add_heralds(
            self.circuit._external_heralds, init_length, final_loc
        )

        # Set axes limits using locations and mode numbers
        self.ax.set_xlim(0, max(self.x_locations) + 0.5)
        self.ax.set_ylim(max(self.y_locations) + 1, -1)
        self.ax.set_yticks(self.y_locations)
        if self.mode_labels is not None:
            exp_len = self.n_modes - len(self.herald_modes)
            if len(self.mode_labels) != exp_len:
                msg = (
                    "Length of provided mode labels list should be equal to "
                    f"the number of useable modes ({exp_len})."
                )
                raise DisplayError(msg)
            mode_labels = self.mode_labels
        else:
            mode_labels = [
                str(i) for i in range(self.n_modes - len(self.herald_modes))
            ]
        mode_labels = [str(m) for m in mode_labels]
        # Include heralded modes in mode labels
        full_mode_labels = []
        count = 0
        for i in range(self.n_modes):
            if i not in self.herald_modes:
                full_mode_labels.append(mode_labels[count])
                count += 1
            else:
                full_mode_labels.append("-")
        self.ax.set_yticklabels(full_mode_labels)
        self.ax.set_xticks([])

        return self.fig, self.ax

    def _add_wg(self, x: float, y: float, length: float) -> None:
        """
        Add a waveguide to the axis.
        """
        # Create rectangle using provided length and waveguide width and then
        # add patch
        rect = patches.Rectangle(
            (x, y - self.wg_width / 2), length, self.wg_width, facecolor="black"
        )
        self.ax.add_patch(rect)

    @multimethod
    def _add(self, spec: Any) -> None:  # noqa: ARG002
        """
        Catch all for any components which may not have been implemented.
        """
        raise ValueError("Unrecognised component in circuit spec.")

    @_add.register
    def _add_ps(self, spec: PhaseShifter) -> None:
        """
        Add a phase shifter to the axis.
        """
        phi = process_parameter_value(spec.phi, self.show_parameter_values)
        # Set size of phase shifter box and length of connector
        size = 0.5
        con_length = 0.5
        # Get x and y locs of target modes
        xloc = self.x_locations[spec.mode]
        yloc = self.y_locations[spec.mode]
        # Add input waveguides
        self._add_wg(xloc, yloc, con_length)
        xloc += con_length
        # Add phase shifter square
        rect = patches.Rectangle(
            (xloc, yloc - size / 2),
            size,
            size,
            facecolor="#e8532b",
            edgecolor="black",
        )
        self.ax.add_patch(rect)
        # Include text to label applied phase shift
        self.ax.text(
            xloc + size / 2,
            yloc,
            "PS",
            horizontalalignment="center",
            verticalalignment="center",
            color="white",
            size=8,
        )
        # Work out value of n*pi/4 closest to phi
        if not isinstance(phi, str):
            n = int(np.round(phi / (np.pi / 4)))
            # Check if value of phi == n*pi/4 to 8 decimal places
            if round(phi, 8) == round(n * np.pi / 4, 8):  # and n > 0:
                n = abs(n)
                # Set text with either pi or pi/2 or pi/4
                if n == 0:
                    phi_text = "0"
                elif n % 4 == 0:
                    phi_text = str(int(n / 4)) + "π" if n > 4 else "π"
                elif n % 4 == 2:
                    phi_text = str(int(n / 2)) + "π/2" if n > 2 else "π/2"
                else:
                    phi_text = str(int(n)) + "π/4" if n > 1 else "π/4"
                if phi < 0:
                    phi_text = "-" + phi_text
            # Otherwise round phi to 4 decimal places
            else:
                phi_text = str(round(phi, 4))
            phi_text = f"$\\phi = {phi_text}$"
        else:
            phi_text = "$\\phi =$ " + phi
        self.ax.text(
            xloc + size / 2,
            yloc + size / 2 + 0.15,
            phi_text,
            color="black",
            size=5,
            horizontalalignment="center",
            verticalalignment="center",
        )
        xloc += size
        # Add output waveguides
        self._add_wg(xloc, yloc, con_length)
        # Update mode locations list
        self.x_locations[spec.mode] = xloc + con_length

    @_add.register
    def _add_bs(self, spec: BeamSplitter) -> None:
        """
        Add a beam splitter across to provided modes to the axis.
        """
        mode1, mode2 = spec.mode_1, spec.mode_2
        ref = process_parameter_value(
            spec.reflectivity, self.show_parameter_values
        )
        if mode1 > mode2:
            mode1, mode2 = mode2, mode1
        size_x = 0.5  # x beam splitter size
        con_length = 0.5  # input/output waveguide length
        offset = 0.5  # Offset of beam splitter shape from mode centres
        size_y = offset + abs(
            self.y_locations[mode2] - self.y_locations[mode1]
        )  # Find y size
        # Get x and y locations
        yloc = self.y_locations[min(mode1, mode2)]
        xloc = max(self.x_locations[mode1 : mode2 + 1])
        # Add initial connectors for any modes which haven't reach xloc yet:
        for i, loc in enumerate(self.x_locations[mode1 : mode2 + 1]):
            if loc < xloc and i + mode1 not in self.herald_modes:
                self._add_wg(loc, self.y_locations[i + mode1], xloc - loc)
        # Add input waveguides for all included modes
        modes = range(min(mode1, mode2), max(mode1, mode2) + 1, 1)
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
        xloc += con_length
        # Add beam splitter rectangle shape
        rect = patches.Rectangle(
            (xloc, yloc - offset / 2),
            size_x,
            size_y,
            facecolor="#3e368d",
            alpha=1,
            edgecolor="black",
        )
        self.ax.add_patch(rect)
        # Label beam splitter as BS and display reflectivity
        self.ax.text(
            xloc + size_x / 2,
            yloc + 0.5,
            "BS",
            size=8,
            horizontalalignment="center",
            verticalalignment="center",
            color="white",
        )
        if not isinstance(ref, str):
            ref = round(ref, 4)
        self.ax.text(
            xloc + size_x / 2,
            yloc + size_y - offset / 2 + 0.15,
            f"$r =$ {ref}",
            color="black",
            size=5,
            horizontalalignment="center",
            verticalalignment="center",
        )
        # For any modes in between the beam splitter modes add a waveguide
        # across the beam splitter
        for i in range(mode1 + 1, mode2):
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], size_x)
        xloc += size_x
        # Add output waveguides and update mode locations
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
            self.x_locations[i] = xloc + con_length

    @_add.register
    def _add_loss(self, spec: Loss) -> None:
        """
        Add a loss channel to the specified mode.
        """
        # Don't add if not enabled
        if not self.display_loss:
            return
        loss = process_parameter_value(spec.loss, self.show_parameter_values)
        # Set size of loss element and input/output waveguide length
        size = 0.5
        con_length = 0.5
        # Get x and y locations
        xloc = self.x_locations[spec.mode]
        yloc = self.y_locations[spec.mode]
        # Add an input waveguide
        self._add_wg(xloc, yloc, con_length)
        xloc += con_length
        # Add loss elements
        rect = patches.Rectangle(
            (xloc, yloc - size / 2),
            size,
            size,
            facecolor="grey",
            edgecolor="black",
        )
        self.ax.add_patch(rect)
        # Label element and add text will loss amount in dB
        self.ax.text(
            xloc + size / 2,
            yloc,
            "L",
            horizontalalignment="center",
            verticalalignment="center",
            color="white",
            size=8,
        )
        if not isinstance(loss, str):
            loss_text = f"$loss = {round(loss, 4)}$"
        else:
            loss_text = "$loss =$ " + loss
        self.ax.text(
            xloc + size / 2,
            yloc + size / 2 + 0.15,
            loss_text,
            color="black",
            size=5,
            horizontalalignment="center",
            verticalalignment="center",
        )
        xloc += size
        # Add output waveguide
        self._add_wg(xloc, yloc, con_length)
        # Update mode position
        self.x_locations[spec.mode] = xloc + con_length

        return

    @_add.register
    def _add_barrier(self, spec: Barrier) -> None:
        """
        Add a barrier which will separate different parts of the circuit. This
        is applied to the provided modes.
        """
        max_loc = 0.0
        for m in spec.modes:
            max_loc = max(max_loc, self.x_locations[m])
        for m in spec.modes:
            loc = self.x_locations[m]
            if loc < max_loc:
                self._add_wg(loc, self.y_locations[m], max_loc - loc)
            self.x_locations[m] = max_loc

    @_add.register
    def _add_mode_swaps(self, spec: ModeSwaps) -> None:
        """
        Add mode swaps between provided modes to the axis.
        """
        # Skip any empty elements
        if not spec.swaps:
            return
        swaps = dict(spec.swaps)  # Make copy of swaps
        size_x = 1  # x beam splitter size
        con_length = 0.25  # input/output waveguide length
        min_mode = min(swaps)
        max_mode = max(swaps)
        # Add in missing mode for swap
        for m in range(min_mode, max_mode + 1):
            if m not in swaps:
                swaps[m] = m
        # Get x and y locations
        xloc = max(self.x_locations[min_mode : max_mode + 1])
        ylocs = []
        for i, j in swaps.items():
            if i not in self.herald_modes:
                ylocs.append((self.y_locations[i], self.y_locations[j]))
        # Add initial connectors for any modes which haven't reach xloc yet:
        for i, loc in enumerate(self.x_locations[min_mode : max_mode + 1]):
            if loc < xloc and i + min_mode not in self.herald_modes:
                self._add_wg(loc, self.y_locations[min_mode + i], xloc - loc)
        # Add input waveguides for all included modes
        modes = range(min_mode, max_mode + 1, 1)
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
        xloc += con_length
        for y0, y1 in ylocs:
            w = self.wg_width / 2
            if y0 < y1:
                dx1 = w * np.arctan(abs(y1 - y0) / size_x)
                dx2 = 0
            else:
                dx1 = 0
                dx2 = w * np.arctan(abs(y1 - y0) / size_x)
            points = [
                (xloc + dx1, y0 - w),
                (xloc, y0 - w),
                (xloc, y0 + w),
                (xloc + dx2, y0 + w),
                (xloc + size_x - dx1, y1 + w),
                (xloc + size_x, y1 + w),
                (xloc + size_x, y1 - w),
                (xloc + size_x - dx2, y1 - w),
            ]
            poly = patches.Polygon(points, facecolor="black")
            self.ax.add_patch(poly)
        xloc += size_x
        # Add output waveguides update mode locations
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
            self.x_locations[i] = xloc + con_length

        return

    @_add.register
    def _add_unitary(self, spec: UnitaryMatrix) -> None:
        """
        Add a unitary representation to the axis.
        """
        mode1, mode2 = spec.mode, spec.mode + spec.unitary.shape[0] - 1
        size_x = 1  # Unitary x size
        con_length = 0.5  # Input/output waveguide lengths
        offset = 0.5  # Offset of unitary square from modes
        size_y = offset + abs(
            self.y_locations[mode2] - self.y_locations[mode1]
        )  # Find total unitary size
        # Get x and y positions
        yloc = self.y_locations[min(mode1, mode2)]
        xloc = max(self.x_locations[mode1 : mode2 + 1])
        # Add initial connectors for any modes which haven't reach xloc yet:
        for i, loc in enumerate(self.x_locations[mode1 : mode2 + 1]):
            if loc < xloc and i + mode1 not in self.herald_modes:
                self._add_wg(loc, self.y_locations[i + mode1], xloc - loc)
        # Add input waveguides
        modes = range(min(mode1, mode2), max(mode1, mode2) + 1, 1)
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
        xloc += con_length
        # Add unitary shape and label
        rect = patches.Rectangle(
            (xloc, yloc - offset / 2),
            size_x,
            size_y,
            facecolor="#1a0f36",
            edgecolor="black",
        )
        self.ax.add_patch(rect)
        s = 10 if len(spec.label) == 1 else 8
        r = 90 if len(spec.label) > 2 else 0
        self.ax.text(
            xloc + size_x / 2,
            yloc + (size_y - offset) / 2,
            spec.label,
            horizontalalignment="center",
            size=s,
            verticalalignment="center",
            color="white",
            rotation=r,
        )
        xloc += size_x
        # Add output waveguides and update mode positions
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(xloc, self.y_locations[i], con_length)
            self.x_locations[i] = xloc + con_length

    @_add.register
    def _add_grouped_circuit(self, spec: Group) -> None:
        """
        Add a grouped circuit drawing to the axis.
        """
        mode1, mode2 = spec.mode_1, spec.mode_2
        if mode1 > mode2:
            mode1, mode2 = mode2, mode1
        size_x = 1  # x size
        con_length = 0.5  # Input/output waveguide lengths
        extra_length = 0.5 if spec.heralds.input or spec.heralds.output else 0
        offset = 0.5  # Offset of square from modes
        size_y = offset + abs(
            self.y_locations[mode2] - self.y_locations[mode1]
        )  # Find total unitary size
        # Get x and y positions
        yloc = self.y_locations[min(mode1, mode2)]
        xloc = max(self.x_locations[mode1 : mode2 + 1])
        # Add initial connectors for any modes which haven't reach xloc yet:
        for i, loc in enumerate(self.x_locations[mode1 : mode2 + 1]):
            if loc < xloc and i + mode1 not in self.herald_modes:
                self._add_wg(loc, self.y_locations[i + mode1], xloc - loc)
        # Add input waveguides
        modes = range(min(mode1, mode2), max(mode1, mode2) + 1, 1)
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(
                    xloc, self.y_locations[i], con_length + extra_length
                )
            elif i - mode1 in spec.heralds.input:
                self._add_wg(
                    xloc + extra_length, self.y_locations[i], con_length
                )
        xloc += con_length + extra_length
        # Add circuit shape and label
        rect = patches.Rectangle(
            (xloc, yloc - offset / 2),
            size_x,
            size_y,
            facecolor="#1a0f36",
            edgecolor="black",
        )
        self.ax.add_patch(rect)
        s = 10 if len(spec.name) == 1 else 8
        r = 90 if len(spec.name) > 2 else 0
        self.ax.text(
            xloc + size_x / 2,
            yloc + (size_y - offset) / 2,
            spec.name,
            horizontalalignment="center",
            size=s,
            verticalalignment="center",
            color="white",
            rotation=r,
        )
        xloc += size_x
        # Add output waveguides and update mode positions
        for i in modes:
            if i not in self.herald_modes:
                self._add_wg(
                    xloc, self.y_locations[i], con_length + extra_length
                )
            elif i - mode1 in spec.heralds.output:
                self._add_wg(xloc, self.y_locations[i], con_length)
            self.x_locations[i] = xloc + con_length + extra_length

        # Modify provided heralds by mode offset and then add at locations
        shifted_heralds = HeraldData(
            input={m + mode1: n for m, n in spec.heralds.input.items()},
            output={m + mode1: n for m, n in spec.heralds.output.items()},
        )
        self._add_heralds(
            shifted_heralds, xloc - size_x - con_length, xloc + con_length
        )

    def _add_heralds(
        self,
        heralds: HeraldData,
        start_loc: float,
        end_loc: float,
    ) -> None:
        """
        Adds display of all heralds to circuit.
        """
        size = 0.2
        # Input heralds
        for mode, num in heralds.input.items():
            xloc = start_loc
            yloc = self.y_locations[mode]
            circle = patches.Circle(
                (xloc, yloc), size, facecolor="#3e368d", edgecolor="black"
            )
            self.ax.add_patch(circle)
            self.ax.text(
                xloc,
                yloc + 0.01,
                str(num),
                color="white",
                size=8,
                horizontalalignment="center",
                verticalalignment="center",
            )
        # Output heralds
        for mode, num in heralds.output.items():
            xloc = end_loc
            yloc = self.y_locations[mode]
            circle = patches.Circle(
                (xloc, yloc), size, facecolor="#3e368d", edgecolor="black"
            )
            self.ax.add_patch(circle)
            self.ax.text(
                xloc,
                yloc + 0.01,
                str(num),
                color="white",
                size=8,
                horizontalalignment="center",
                verticalalignment="center",
            )



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from lightworks.qubit.gates.single_qubit_gates import (
    SX,
    H,
    P,
    Rx,
    Ry,
    Rz,
    S,
    Sadj,
    T,
    Tadj,
    X,
    Y,
    Z,
)
from lightworks.qubit.gates.three_qubit_gates import CCNOT, CCZ
from lightworks.qubit.gates.two_qubit_gates import (
    CNOT,
    CZ,
    SWAP,
    CNOT_Heralded,
    CZ_Heralded,
)
from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.utils.exceptions import LightworksError
from lightworks.sdk.utils.post_selection import PostSelection

from . import QISKIT_INSTALLED

if QISKIT_INSTALLED:
    from qiskit import QuantumCircuit

SINGLE_QUBIT_GATES_MAP = {
    "h": H(),
    "x": X(),
    "y": Y(),
    "z": Z(),
    "s": S(),
    "sdg": Sadj(),
    "t": T(),
    "tdg": Tadj(),
    "sx": SX(),
}
ROTATION_GATES_MAP = {"rx": Rx, "ry": Ry, "rz": Rz, "p": P}

TWO_QUBIT_GATES_MAP = {"cx": CNOT_Heralded, "cz": CZ_Heralded, "swap": SWAP}
TWO_QUBIT_GATES_MAP_PS = {"cx": CNOT, "cz": CZ}

THREE_QUBIT_GATES_MAP = {"ccx": CCNOT, "ccz": CCZ}

ALLOWED_GATES = [
    *SINGLE_QUBIT_GATES_MAP,
    *ROTATION_GATES_MAP,
    *TWO_QUBIT_GATES_MAP,
    *THREE_QUBIT_GATES_MAP,
]


def qiskit_converter(
    circuit: "QuantumCircuit", allow_post_selection: bool = False
) -> tuple[PhotonicCircuit, PostSelection | None]:
    """
    Performs conversion of a provided qiskit QuantumCircuit into a photonic
    circuit within Lightworks.

    Args:

        circuit (QuantumCircuit) : The qiskit circuit to be converted.

        allow_post_selection (bool, optional) : Controls whether post-selected
            gates can be utilised within the circuit.

    Returns:

        PhotonicCircuit : The created circuit within Lightworks.

        PostSelection | None : If post-selection rules are required for the
            created circuit, then an object which implements these will be
            returned, otherwise it will be None.

    """
    converter = QiskitConverter(allow_post_selection)
    return converter.convert(circuit)


class QiskitConverter:
    """
    Manages conversion between qiskit and lightworks circuit, adding each of the
    qubit gates into a created circuit object.

    Args:

        allow_post_selection (bool, optional) : Controls whether post-selected
            gates can be utilised within the circuit.

    """

    def __init__(self, allow_post_selection: bool = False) -> None:
        self.allow_post_selection = allow_post_selection

    def convert(
        self, q_circuit: "QuantumCircuit"
    ) -> tuple[PhotonicCircuit, PostSelection | None]:
        """
        Performs conversion of a provided qiskit QuantumCircuit into a photonic
        circuit within Lightworks.

        Args:

            q_circuit (QuantumCircuit) : The qiskit circuit to be converted.

        Returns:

            PhotonicCircuit : The created circuit within Lightworks.

            PostSelection | None : If post-selection rules are required for the
                created circuit, then an object which implements these will be
                returned, otherwise it will be None.

        """
        if not QISKIT_INSTALLED:
            raise LightworksError(
                "Lightworks qiskit optional requirements not installed, "
                "this can be achieved with 'pip install lightworks[qiskit]'."
            )

        if not isinstance(q_circuit, QuantumCircuit):
            raise TypeError(
                "PhotonicCircuit to convert must be a qiskit circuit."
            )

        n_qubits = q_circuit.num_qubits
        self.circuit = PhotonicCircuit(n_qubits * 2)
        self.modes = {i: (2 * i, 2 * i + 1) for i in range(n_qubits)}

        if self.allow_post_selection:
            post_select, ps_qubits = post_selection_analyzer(q_circuit)
        else:
            post_select = [False] * len(q_circuit.data)

        for i, inst in enumerate(q_circuit.data):
            gate = inst.operation.name
            qubits = [
                inst.qubits[i]._index for i in range(inst.operation.num_qubits)
            ]
            if gate not in ALLOWED_GATES:
                msg = f"Unsupported gate '{gate}' included in circuit."
                raise ValueError(msg)
            # Single Qubit Gates
            if len(qubits) == 1:
                if gate in SINGLE_QUBIT_GATES_MAP:
                    self._add_single_qubit_gate(gate, *qubits)
                else:
                    theta = inst.operation.params[0]
                    self._add_single_qubit_rotation_gate(gate, theta, *qubits)
            # Two Qubit Gates
            elif len(qubits) == 2:
                self._add_two_qubit_gate(
                    gate,
                    *qubits,  # type: ignore[call-arg]
                    post_select[i],
                )
            # Three Qubit Gates
            elif len(qubits) == 3:
                self._add_three_qubit_gate(
                    gate,
                    *qubits,  # type: ignore[call-arg]
                    post_select[i],
                )
            # Limit to three qubit gates
            else:
                raise ValueError("Gates with more than 3 qubits not supported.")

        if self.allow_post_selection:
            if ps_qubits:
                ps_rules = PostSelection()
                for q in ps_qubits:
                    ps_rules.add(self.modes[q], 1)
            else:
                ps_rules = None
        else:
            ps_rules = None

        return (self.circuit, ps_rules)

    def _add_single_qubit_gate(self, gate: str, qubit: int) -> None:
        """
        Adds a single qubit gate to the selected qubit on the circuit.
        """
        self.circuit.add(SINGLE_QUBIT_GATES_MAP[gate], self.modes[qubit][0])

    def _add_single_qubit_rotation_gate(
        self, gate: str, theta: float, qubit: int
    ) -> None:
        """
        Adds a single qubit gate to the selected qubit on the circuit.
        """
        self.circuit.add(ROTATION_GATES_MAP[gate](theta), self.modes[qubit][0])

    def _add_two_qubit_gate(
        self, gate: str, q0: int, q1: int, post_selection: bool = False
    ) -> None:
        """
        Adds a two qubit gate to the circuit on the selected qubits.
        """
        if gate == "swap":
            self.circuit.add(
                TWO_QUBIT_GATES_MAP["swap"](self.modes[q0], self.modes[q1]), 0
            )
        elif gate in {"cx", "cz"}:
            mapper = (
                TWO_QUBIT_GATES_MAP
                if not post_selection
                else TWO_QUBIT_GATES_MAP_PS
            )
            q0, q1, to_swap = convert_two_qubits_to_adjacent(q0, q1)
            if gate == "cx":
                target = q1 - min([q0, q1])
                add_circ = mapper["cx"](target)
            else:
                add_circ = mapper["cz"]()
            add_mode = self.modes[min([q0, q1])][0]
            for swap_qs in to_swap:
                self._add_two_qubit_gate("swap", swap_qs[0], swap_qs[1])
            self.circuit.add(add_circ, add_mode)
            for swap_qs in to_swap:
                self._add_two_qubit_gate("swap", swap_qs[0], swap_qs[1])
        else:
            msg = f"Unsupported gate '{gate}' included in circuit."
            raise ValueError(msg)

    def _add_three_qubit_gate(
        self, gate: str, q0: int, q1: int, q2: int, post_selection: bool = False
    ) -> None:
        """
        Adds a three qubit gate to the circuit on the selected qubits.
        """
        if gate in {"ccx", "ccz"}:
            if not post_selection:
                raise ValueError(
                    "Three qubit gates can only be used with post-selection. "
                    "Ensure allow_post_selection is True to enable this. The "
                    "location of the gate may also need to be towards the end "
                    "of the circuit as a result of requirements on "
                    "post-selection.."
                )
            all_qubits = [q0, q1, q2]
            if max(all_qubits) - min(all_qubits) != 2:
                raise ValueError(
                    "CCX and CCZ qubits must be adjacent to each other, "
                    "please add swap gates to achieve this."
                )
            if gate == "ccx":
                target = q2 - min(all_qubits)
                add_circ = THREE_QUBIT_GATES_MAP["ccx"](target)
            else:
                add_circ = THREE_QUBIT_GATES_MAP["ccz"]()
            add_mode = self.modes[min(all_qubits)][0]
            self.circuit.add(add_circ, add_mode)
        else:
            msg = f"Unsupported gate '{gate}' included in circuit."
            raise ValueError(msg)


def convert_two_qubits_to_adjacent(
    q0: int, q1: int
) -> tuple[int, int, list[tuple[int, int]]]:
    """
    Takes two qubit indices and converts these so that they are adjacent to each
    other, and determining the swaps required for this. The order of the two
    qubits is preserved, so if q0 > q1 then this will remain True.

    Args:

        q0 (int) : First qubit which a gate acts on.

        q1 (int) : The second qubit which the gate acts on.

    Returns:

        int : The new first qubit which the gate should act on.

        int : The new second qubit which the gate should act on.

        list[tuple] : Pairs of qubits which swap gates should be applied to
            ensure the gate can act on the right qubits.

    """
    if abs(q1 - q0) == 1:
        return (q0, q1, [])
    swaps = []
    new_upper = max(q0, q1)
    new_lower = min(q0, q1)
    # Bring modes closer together until they are adjacent
    while new_upper - new_lower != 1:
        new_upper -= 1
        if new_upper - new_lower == 1:
            break
        new_lower += 1
    if min(q0, q1) != new_lower:
        swaps.append((min(q0, q1), new_lower))
    if max(q0, q1) != new_upper:
        swaps.append((max(q0, q1), new_upper))
    if q0 < q1:
        q0, q1 = new_lower, new_upper
    else:
        q0, q1 = new_upper, new_lower
    return (q0, q1, swaps)


def post_selection_analyzer(
    qc: "QuantumCircuit",
) -> tuple[list[bool], list[int]]:
    """
    Implements a basic algorithm to try to determine which gates can have
    post-selection and which require heralding. This is not necessarily optimal,
    but should at least reduce heralding requirements.

    Args:

        qc (QuantumCircuit) : The qiskit circuit to be analysed.

    Returns:

        list[bool] : A list of length elements in the circuit that indicates if
            each element is compatible with post-selection. This will include
            any single qubit gates, even though post-selection is not relevant
            here.

        list[int] : A list of integers indicating which qubits need a
            post-selection rule to be applied.

    """
    # First extract all qubit data from the circuit
    gate_qubits: list[list[int] | None] = []
    for inst in qc.data:
        if inst.operation.num_qubits >= 2:
            gate_qubits.append(
                [
                    inst.qubits[i]._index
                    for i in range(inst.operation.num_qubits)
                ]
            )
        else:
            gate_qubits.append(None)

    post_selection = []
    has_ps = []
    # Work backwards through gates
    for gate in reversed(gate_qubits):
        if gate is None:
            post_selection.append(False)
            continue
        can_ps = not all(q in has_ps for q in gate)
        post_selection.append(can_ps)
        has_ps += gate
    # Return if a gate can have post-selection and all modes which will require
    # it.
    return list(reversed(post_selection)), list(set(has_ps))



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from types import NoneType

from lightworks.emulator.components import Detector, Source
from lightworks.sdk.circuit import PhotonicCircuit
from lightworks.sdk.results import ProbabilityDistribution
from lightworks.sdk.state import State
from lightworks.sdk.utils.exceptions import ModeMismatchError, SamplerError
from lightworks.sdk.utils.post_selection import (
    PostSelectionType,
    process_post_selection,
)
from lightworks.sdk.utils.random import process_random_seed
from lightworks.sdk.utils.state import check_herald_difference

from .data import SamplerTask
from .task import Task


class Sampler(Task):
    """
    Used to randomly sample from the photon number distribution output of a
    provided circuit. The distribution is calculated when the class is first
    called and then used to return output states with the sample function. Also
    supports the inclusion of imperfect source and detection properties.

    Args:

        circuit (PhotonicCircuit) : The circuit to sample output states from.

        input_state (State) : The input state to use with the circuit for
            sampling.

        n_samples (int) : The number of samples that are to be returned.

        source (Source, optional) : Provide a source object to simulate an
            imperfect input. This defaults to None, which will create a perfect
            source.

        detector (Detector, optional) : Provide detector to simulate imperfect
            detector probabilities. This defaults to None, which will assume a
            perfect detector.

        post_selection (PostSelection | function, optional) : A PostSelection
            object or function which applies a provided set of post-selection
            criteria to a state.

        min_detection (int | None, optional) : Post-select on a given minimum
            total number of photons, this should not include any heralded
            photons. If set to None then this will be the smallest valuable
            possible on a platform.

        random_seed (int|None, optional) : Option to provide a random seed to
            reproducibly generate samples from the function. This is
            optional and can remain as None if this is not required.

        sampling_mode (str, optional) : Sets the mode of the Sampler. In input
            mode, N cycles of the system are measured, and only results which
            meet any assigned criteria are returned. In output mode, N valid
            samples are produced from the system. Should be either 'input' or
            'output', defaults to 'output'.

    """

    def __init__(
        self,
        circuit: PhotonicCircuit,
        input_state: State,
        n_samples: int,
        post_selection: PostSelectionType
        | Callable[[State], bool]
        | None = None,
        min_detection: int | None = None,
        source: Source | None = None,
        detector: Detector | None = None,
        random_seed: int | None = None,
        sampling_mode: str = "output",
    ) -> None:
        # Assign provided quantities to attributes
        self.circuit = circuit
        self.input_state = input_state
        self.source = source
        self.detector = detector
        self.n_samples = n_samples
        self.post_selection = post_selection  # type: ignore[assignment]
        self.min_detection = min_detection
        self.random_seed = random_seed
        self.sampling_mode = sampling_mode
        self._probability_distribution: ProbabilityDistribution

    @property
    def circuit(self) -> PhotonicCircuit:
        """
        Stores the circuit to be used for simulation, should be a
        PhotonicCircuit object.
        """
        return self.__circuit

    @circuit.setter
    def circuit(self, value: PhotonicCircuit) -> None:
        if not isinstance(value, PhotonicCircuit):
            raise TypeError(
                "Provided circuit should be a PhotonicCircuit or Unitary "
                "object."
            )
        self.__circuit = value

    @property
    def input_state(self) -> State:
        """The input state to be used for sampling."""
        return self.__input_state

    @input_state.setter
    def input_state(self, value: State) -> None:
        if not isinstance(value, State):
            raise TypeError("A single input of type State should be provided.")
        if len(value) != self.circuit.input_modes:
            raise ModeMismatchError(
                "Incorrect input length for provided circuit."
            )
        # Also validate state values
        value._validate()
        self.__input_state = value

    @property
    def source(self) -> Source | None:
        """
        Details the properties of the Source used for creation of inputs to
        the Sampler.
        """
        return self.__source

    @source.setter
    def source(self, value: Source | None) -> None:
        if not isinstance(value, Source | NoneType):
            raise TypeError("Provided source should be a Source object.")
        self.__source = value

    @property
    def detector(self) -> Detector | None:
        """
        Details the properties of the Detector used for photon measurement.
        """
        return self.__detector

    @detector.setter
    def detector(self, value: Detector | None) -> None:
        if not isinstance(value, Detector | NoneType):
            raise TypeError("Provided detector should be a Detector object.")
        self.__detector = value

    @property
    def n_samples(self) -> int:
        """Stores the number of samples to be collected in an experiment."""
        return self.__n_samples

    @n_samples.setter
    def n_samples(self, value: int) -> None:
        if not isinstance(value, int) or isinstance(value, bool):
            raise TypeError("n_samples must be an integer")
        if value < 0:
            raise ValueError("Number of samples should be larger than 0.")
        self.__n_samples = value

    @property
    def post_selection(self) -> PostSelectionType | None:
        """Describes the post-selection criteria to be applied to a state."""
        return self.__post_selection

    @post_selection.setter
    def post_selection(
        self, value: PostSelectionType | Callable[[State], bool] | None
    ) -> None:
        self.__post_selection = process_post_selection(value)

    @property
    def min_detection(self) -> int | None:
        """
        Stores the minimum number of photons to be measured in an experiment,
        this excludes heralded photons.
        """
        return self.__min_detection

    @min_detection.setter
    def min_detection(self, value: int | None) -> None:
        if not isinstance(value, int | NoneType) or isinstance(value, bool):
            raise TypeError("min_detection must be an integer or None.")
        self.__min_detection = value

    @property
    def random_seed(self) -> int | None:
        """
        Stores a random seed which is used for gathering repeatable data from
        the Sampler
        """
        return self.__random_seed

    @random_seed.setter
    def random_seed(self, value: int | None) -> None:
        self.__random_seed = process_random_seed(value)

    @property
    def sampling_mode(self) -> str:
        """
        Stores the input mode of the sampler, which controls whether N valid
        inputs or outputs are produced from the system.
        """
        return self.__sampling_mode

    @sampling_mode.setter
    def sampling_mode(self, value: str) -> None:
        if value not in {"input", "output"}:
            raise ValueError(
                "Sampling mode must be set to either input or output."
            )
        self.__sampling_mode = value

    @property
    def probability_distribution(self) -> ProbabilityDistribution:
        """
        Returns the calculated probability distribution. This can only be called
        once the Sampler has been run on the backend. Note: If any changes are
        made to the Sampler configuration, it should be re-evaluated with the
        backend to update this distribution.
        """
        if not hasattr(self, "_probability_distribution"):
            raise SamplerError(
                "Probability distribution must be calculated by first running "
                "the Sampler on a backend."
            )
        return self._probability_distribution

    def _generate_task(self) -> SamplerTask:
        check_herald_difference(self.circuit, self.input_state.n_photons)
        return SamplerTask(
            circuit=self.circuit._build(),
            input_state=self.input_state,
            n_samples=self.n_samples,
            source=self.source,
            detector=self.detector,
            post_selection=self.post_selection,
            min_detection=self.min_detection,
            random_seed=self.random_seed,
            sampling_mode=self.sampling_mode,
        )



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numpy.typing import NDArray

from lightworks.sdk.state import State

from .mappings import PAULI_MAPPING, RHO_MAPPING
from .process_tomography import _ProcessTomography
from .utils import (
    _calculate_expectation_value,
    _combine_all,
    _unvec,
    _vec,
    process_fidelity,
)


class LIProcessTomography(_ProcessTomography):
    """
    Runs quantum process tomography using the linear inversion estimation
    method.

    Args:

        n_qubits (int) : The number of qubits that will be used as part of the
            tomography.

        base_circuit (PhotonicCircuit) : An initial circuit which realises the
            required operation and can be modified for performing tomography.
            It is required that the number of circuit input modes equals 2 * the
            number of qubits.

    """

    @property
    def choi(self) -> NDArray[np.complex128]:
        """Returns the calculate choi matrix for a circuit."""
        if not hasattr(self, "_choi"):
            raise AttributeError(
                "Choi matrix has not yet been calculated, this can be achieved "
                "with the process method."
            )
        return self._choi

    def process(
        self,
        data: list[dict[State, int]] | dict[tuple[str, str], dict[State, int]],
    ) -> NDArray[np.complex128]:
        """
        Performs process tomography with the configured elements and calculates
        the choi matrix using linear inversion.

        Args:

            data (list | dict) : The collected measurement data. If a list then
                this should match the order the experiments were provided, and
                if a dictionary, then each key should be tuple of the input and
                measurement basis.

        Returns:

            np.ndarray : The calculated choi matrix for the process.

        """
        results = self._convert_tomography_data(data)
        # Get expectation values using results
        lambdas = self._calculate_expectation_values(results)
        # Find all pauli and density matrices for multi-qubit states
        full_paulis = _combine_all(PAULI_MAPPING, self.n_qubits)
        full_rhos = _combine_all(RHO_MAPPING, self.n_qubits)
        # Determine the transformation matrix to perform linear inversion
        dim = 2**self.n_qubits
        transform_matrix = np.zeros((dim**4, dim**4), dtype=complex)
        for i, (in_s, meas) in enumerate(lambdas):
            transform_matrix[i, :] = _vec(
                np.kron(np.array(full_rhos[in_s]).conj(), full_paulis[meas])
            ).conj()
        # Then find the choi matrix
        choi = np.linalg.pinv(transform_matrix) @ np.array(
            list(lambdas.values())
        )
        self._choi = _unvec(choi)
        return self.choi

    def fidelity(self, choi_exp: NDArray[np.complex128]) -> float:
        """
        Calculates fidelity of the calculated choi matrix compared to the
        expected one.
        """
        return process_fidelity(self.choi, choi_exp)

    def _calculate_expectation_values(
        self, results: dict[tuple[str, str], dict[State, int]]
    ) -> dict[tuple[str, str], float]:
        """
        Calculates the expectation values from a set of results containing,
        input states, observables and measurement data for each.
        """
        return {
            (in_state, measurement): _calculate_expectation_value(
                measurement, res
            )
            for (in_state, measurement), res in results.items()
        }



# Copyright 2024 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, TypeVar

import numpy as np
from multimethod import multimethod
from numpy.typing import NDArray
from scipy.linalg import sqrtm

from lightworks.sdk.state import State

from .mappings import MEASUREMENT_MAPPING, PAULI_MAPPING

# NOTE: This file is auto-documented, so any functions not intended to be public
# should begin with _

T = TypeVar("T", bound=np.generic)


class TomographyDataError(Exception):
    """
    Raised when required data is missing for performing a tomography algorithm.
    """


def state_fidelity(
    rho: NDArray[np.complex128], rho_exp: NDArray[np.complex128]
) -> float:
    """
    Calculates the fidelity of the density matrix for a quantum state against
    the expected density matrix.

    Args:

        rho (np.ndarray) : The calculated density matrix of the quantum state.

        rho_exp (np.ndarray) : The expected density matrix.

    Returns:

        float : The calculated fidelity value.

    """
    rho_exp = np.array(rho_exp)
    rho_root = sqrtm(np.array(rho))
    if rho_root.shape != rho_exp.shape:
        msg = (
            "Mismatch in dimensions between provided density matrices, "
            f"{rho_root.shape} & {rho_exp.shape}."
        )
        raise ValueError(msg)
    inner = rho_root @ rho_exp @ rho_root
    return np.real(np.trace(sqrtm(inner)) ** 2)


def process_fidelity(
    choi: NDArray[np.complex128], choi_exp: NDArray[np.complex128]
) -> float:
    """
    Calculates the fidelity of a process compared to an expected choi matrix.

    Args:

        choi (np.ndarray) : The calculated choi matrix for the process.

        choi_exp (np.ndarray) : The expected choi matrix.

    Returns:

        float : The calculated fidelity value.

    """
    if choi.shape != choi_exp.shape:
        msg = (
            "Mismatch in dimensions between provided density matrices, "
            f"{choi.shape} & {choi_exp.shape}."
        )
        raise ValueError(msg)
    n_qubits = int(np.emath.logn(4, choi.shape[0]))
    return state_fidelity(choi / 2**n_qubits, choi_exp / 2**n_qubits)


def density_from_state(
    state: list[complex] | NDArray[np.complex128],
) -> NDArray[np.complex128]:
    """
    Calculates the expected density matrix from a given state.

    Args:

        state (list | np.ndarray) : The vector representation of the state for
            which the density matrix should be calculated.

    Returns:

        np.ndarray : The calculated density matrix.

    """
    state = np.array(state)
    return np.outer(state, np.conj(state.T)).astype(complex)


def choi_from_unitary(
    unitary: NDArray[np.complex128],
) -> NDArray[np.complex128]:
    """
    Calculates the expected choi matrix from a given unitary representation of a
    process.

    Args:

        unitary (np.ndarray) : The unitary representation of the gate.

    Returns:

        np.ndarray : The calculated choi matrix.

    """
    unitary = np.array(unitary)
    return np.outer(unitary.flatten(), np.conj(unitary.flatten())).astype(
        complex
    )


def _vec(mat: NDArray[T]) -> NDArray[T]:
    """
    Applies flatten operation to a provided matrix to convert it into a vector.
    """
    return mat.flatten()


def _unvec(mat: NDArray[T]) -> NDArray[T]:
    """
    Takes a provided vector and converts it into a square matrix.
    """
    dim = int(mat.shape[0] ** 0.5)
    return mat.reshape(dim, dim)


@multimethod
def _combine_all(value: Any, n: int) -> None:  # noqa: ARG001
    """
    Combines all elements of provided value with itself n number of times.
    """
    raise TypeError("combine_all method not implemented for provided type.")


@_combine_all.register
def _combine_all_list(value: list[str], n: int) -> list[str]:
    """
    Sums string values within list.
    """
    result = list(value)
    for _ in range(n - 1):
        result = [v1 + "," + v2 for v1 in result for v2 in value]
    return result


@_combine_all.register
def _combine_all_list_array(
    value: list[np.ndarray[Any, Any]], n: int
) -> list[NDArray[Any]]:
    """
    Performs tensor product of all combinations of arrays within list.
    """
    result = list(value)
    for _ in range(n - 1):
        result = [np.kron(v1, v2) for v1 in result for v2 in value]
    return result


@_combine_all.register
def _combine_all_dict_mat(
    value: dict[str, np.ndarray[Any, Any]], n: int
) -> dict[str, NDArray[Any]]:
    """
    Sums keys of dictionary and performs tensor products of the dictionary
    values.
    """
    result = dict(value)
    for _ in range(n - 1):
        result = {
            k1 + "," + k2: np.kron(v1, v2)
            for k1, v1 in result.items()
            for k2, v2 in value.items()
        }
    return result


def _get_tomo_measurements(
    n_qubits: int, remove_trivial: bool = False
) -> list[str]:
    """
    Returns all measurements required for a state tomography of n qubits.

    Args:

        n_qubits (int) : The number of qubits used in the tomography.

        remove_trivial (bool) : Allows for removal of the trivial I*n_qubits
            measurement when this is not required

    Returns:

        list : A list of the measurement combinations for tomography.

    """
    all_meas = _combine_all(list(MEASUREMENT_MAPPING.keys()), n_qubits)
    if remove_trivial:
        all_meas.pop(all_meas.index(",".join("I" * n_qubits)))
    return all_meas


def _get_required_tomo_measurements(
    n_qubits: int, remove_trivial: bool = False
) -> tuple[list[str], dict[str, str]]:
    """
    Calculates reduced list of required measurements assuming that any
    measurements in the I basis can be replaced with a Z measurement.
    A dictionary which maps the full measurements to the reduced basis is
    also returned.

    Args:

        n_qubits (int) : The number of qubits used in the tomography.

        remove_trivial (bool) : Allows for removal of the trivial I*n_qubits
            measurement when this is not required

    Returns:

        list : A list of the minimum required measurement combinations for
            tomography.

        dict : A mapping between the full set of measurement operators and
            the required minimum set.

    """
    mapping = {
        c: c.replace("I", "Z")
        for c in _get_tomo_measurements(n_qubits, remove_trivial=remove_trivial)
    }
    req_measurements = list(set(mapping.values()))
    req_measurements.sort()
    return req_measurements, mapping


def _calculate_expectation_value(
    measurement: str, results: dict[State, int]
) -> float:
    """
    Calculates the expectation value for a given measurement and set of
    results.

    Args:

        measurement (str) : The measurement operator used for the
            computation.

        results (dict) : A dictionary of measured output states and counts.

    Returns:

        float : The calculated expectation value.

    """
    expectation = 0
    n_counts = 0
    for state, counts in results.items():
        n_counts += counts
        # Adjust multiplier to account for variation in eigenvalues
        multiplier = 1
        for j, gate in enumerate(measurement.split(",")):
            if gate == "I" or state[2 * j : 2 * j + 2] == State([1, 0]):
                multiplier *= 1
            elif state[2 * j : 2 * j + 2] == State([0, 1]):
                multiplier *= -1
            else:
                msg = (
                    f"An invalid state {state[2 * j : 2 * j + 2]} was found"
                    " in the results. This does not correspond to a valid "
                    "value for dual-rail encoded qubits."
                )
                raise ValueError(msg)
        expectation += multiplier * counts
    return expectation / n_counts


def _calculate_density_matrix(
    results: dict[str, dict[State, int]], n_qubits: int
) -> NDArray[np.complex128]:
    """
    Calculates the density matrix using a provided dictionary of results
    data.

    Args:

        results (dict) : Contains data of outputs and counts for each
            corresponding set of measurement indices.

        n_qubits (int) : The number of qubits in the experiment.

    Returns:

        np.ndarray : The calculated density matrix.


    """
    # Process results to find density matrix
    rho = np.zeros((2**n_qubits, 2**n_qubits), dtype=complex)
    for measurement, result in results.items():
        expectation = _calculate_expectation_value(measurement, result)
        expectation /= 2**n_qubits
        # Calculate tensor product of the operators used
        ops = measurement.split(",")
        mat = PAULI_MAPPING[ops[0]]
        for g in ops[1:]:
            mat = np.kron(mat, PAULI_MAPPING[g]).astype(complex)
        # Updated density matrix
        rho += expectation * mat
    return rho


def _check_target_process(
    process_mat: NDArray[np.complex128], n_qubits: int
) -> NDArray[np.complex128]:
    """
    Confirms the target process matrix is a numpy array of the correct shape.
    """
    try:
        process_mat = np.array(process_mat, dtype=complex)
    except Exception as e:
        raise TypeError(
            "Provided process matrix could not be converted to a complex numpy "
            "array."
        ) from e
    if process_mat.shape != (2**n_qubits, 2**n_qubits):
        msg = (
            "Provided process matrix has incorrect shape, expected "
            f"{(2**n_qubits, 2**n_qubits)}, got {process_mat.shape}."
        )
        raise ValueError(msg)
    return process_mat
